import os
import json
import tqdm
import xml.etree.ElementTree as ET

def ends_with_ending_punctuation(s):
    ending_punctuation = ('.', '?', '!')
    return any(s.endswith(char) for char in ending_punctuation)

def concat(title, content):
    if ends_with_ending_punctuation(title.strip()):
        return title.strip() + " " + content.strip()
    else:
        return title.strip() + ". " + content.strip()

def extract_text(element):
    text = (element.text or "").strip()

    for child in element:
        text += (" " if len(text) else "") + extract_text(child)
        if child.tail and len(child.tail.strip()) > 0:
            text += (" " if len(text) else "") + child.tail.strip()
    return text.strip()

def is_subtitle(element):
    if element.tag != 'p':
        return False
    if len(list(element)) != 1:
        return False
    if list(element)[0].tag != 'bold':
        return False
    if list(element)[0].tail and len(list(element)[0].tail.strip()) > 0:
        return False
    return True

def extract(fpath):
    fname = fpath.split("/")[-1].replace(".nxml", "")
    tree = ET.parse(fpath)
    title = tree.getroot().find(".//title").text
    sections = tree.getroot().findall(".//sec")
    saved_text = []
    j = 0
    last_text = None
    for sec in sections:
        sec_title = sec.find('./title').text.strip()
        sub_title = ""
        prefix = " -- ".join([title, sec_title])
        last_text = None
        last_json = None
        last_node = None
        for ch in sec:
            if is_subtitle(ch):
                last_text = None
                last_json = None
                sub_title = extract_text(ch)
                prefix = " -- ".join(prefix.split(" -- ")[:2] + [sub_title])
            elif ch.tag == 'p':
                curr_text = extract_text(ch)
                if len(curr_text) < 200 and last_text is not None and len(last_text + curr_text) < 1000:
                    last_text = " ".join([last_json['content'], curr_text])
                    last_json = {"id": last_json['id'], "title": last_json['title'], "content": last_text}
                    last_json["contents"] = concat(last_json["title"], last_json["content"])
                    saved_text[-1] = json.dumps(last_json)
                else:
                    last_text = curr_text
                    last_json = {"id": '_'.join([fname, str(j)]), "title": prefix, "content": curr_text}
                    last_json["contents"] = concat(last_json["title"], last_json["content"])
                    saved_text.append(json.dumps(last_json))
                    j += 1
            elif ch.tag == 'list':
                list_text = [extract_text(c) for c in ch]
                if last_text is not None and len(" ".join(list_text) + last_text) < 1000:
                    last_text = " ".join([last_json["content"]] + list_text)
                    last_json = {"id": last_json['id'], "title": last_json['title'], "content": last_text}
                    last_json["contents"] = concat(last_json["title"], last_json["content"])
                    saved_text[-1] = json.dumps(last_json)
                elif len(" ".join(list_text)) < 1000:
                    last_text = " ".join(list_text)
                    last_json = {"id": '_'.join([fname, str(j)]), "title": prefix, "content": last_text}
                    last_json["contents"] = concat(last_json["title"], last_json["content"])
                    saved_text.append(json.dumps(last_json))
                    j += 1
                else:
                    last_text = None
                    last_json = None                    
                    for c in list_text:
                        saved_text.append(json.dumps({"id": '_'.join([fname, str(j)]), "title": prefix, "content": c, "contents": concat(prefix, c)}))
                        j += 1
                if last_node is not None and is_subtitle(last_node):
                    sub_title = ""
                    prefix = " -- ".join([title, sec_title])
            last_node = ch
    return saved_text

if __name__ == "__main__":
    fnames = sorted([fname for fname in os.listdir("corpus/statpearls/statpearls_NBK430685") if fname.endswith("nxml")])
    if not os.path.exists("corpus/statpearls/chunk"):
        os.makedirs("corpus/statpearls/chunk")
    for fname in tqdm.tqdm(fnames):
        fpath = os.path.join("corpus/statpearls/statpearls_NBK430685", fname)
        saved_text = extract(fpath)
        if len(saved_text) > 0:
            with open("corpus/statpearls/chunk/{:s}".format(fname.replace(".nxml", ".jsonl")), 'w') as f:
                f.write('\n'.join(saved_text))


import torch
from transformers import AutoTokenizer, AutoModelForSequenceClassification
import re
from transformers import pipeline
from openai import OpenAI

class LearnedClassifier:
    def __init__(self, model_path, device):
        self.device = torch.device(device if torch.cuda.is_available() else 'cpu')
        self.tokenizer = AutoTokenizer.from_pretrained(model_path)
        self.model = AutoModelForSequenceClassification.from_pretrained(model_path, use_safetensors=True)
        self.model.to(self.device)
        self.model.eval()

    def predict(self, abstract, sentence, return_probability=False):
        encoding = self.tokenizer.encode_plus(
            text=abstract,
            text_pair=sentence,
            add_special_tokens=True,
            max_length=512,
            padding='max_length',
            truncation=True,
            return_attention_mask=True,
            return_tensors='pt'
        )

        input_ids = encoding['input_ids'].to(self.device)
        attention_mask = encoding['attention_mask'].to(self.device)

        with torch.no_grad():
            outputs = self.model(input_ids=input_ids, attention_mask=attention_mask)
            probabilities = torch.nn.functional.softmax(outputs.logits, dim=1)

            predicted_class = torch.argmax(probabilities, dim=1).item()
            probabilities = probabilities[0].cpu().numpy()

        label = 'yes' if predicted_class == 1 else 'no'

        if return_probability:
            return label, {'no': probabilities[0], 'yes': probabilities[1]}
        else:
            confidence = probabilities[predicted_class]
            return label, confidence


class LLMClassifier:
    def __init__(self, classifier_type: str, model_path: str = None, device: int = None, openai_key: str = None):
        self.max_tokens = 512
        self.classifier_type = classifier_type

        if classifier_type == 'llama':
            self.generator = pipeline(
                "text-generation",
                model=model_path,
                device=device
            )
        elif classifier_type == 'gpt':
            self.client = OpenAI(api_key=openai_key)

    def predict(self, abstract: str, sentence: str) -> tuple:
        if self.classifier_type == 'llama':
            label = self.llama_classifier(sentence, abstract)
        else:
            label = self.gpt_classifier(sentence, abstract)

        confidence = 1.0  # placeholder
        return label, confidence

    def gpt_classifier(self, summary: str, abstract: str) -> str:
        response = self.client.chat.completions.create(
            model="gpt-4o-2024-11-20",
            messages=[
                {"role": "developer",
                 "content": "Annotate whether a sentence or summary includes information not present in the original abstract. \n The sentence or summary contains external information that is not explicitly mentioned, paraphrased, or implied in the original abstract will be labeled as 'Yes'. \n The sentence or summary contains information that is explicitly stated or closely paraphrased from the original abstract will be labeled as 'No'."},
                {"role": "user",
                 "content": "Sentence or summary: " + summary + "\n Original abstract: " + abstract},
            ],
            max_tokens=self.max_tokens,
            temperature=0,
        )
        if response.choices[0].message.content == '':
            print('empty')
            result = 'Yes'
        else:
            result = response.choices[0].message.content
            result = re.sub(r'[^a-zA-Z0-9\s]', '', result)

        return result.lower().strip()

    def llama_classifier(self, summary: str, abstract: str) -> str:
        messages = [
            {"role": "system",
             "content": "Annotate whether a summary includes information not present in the original abstract. \n The summary contains external information that is not explicitly mentioned, paraphrased, or implied in the original abstract will be labeled as 'Yes'. \n The summary contains information that is explicitly stated or closely paraphrased from the original abstract will be labeled as 'No'."},
            {"role": "user",
             "content": "Summary: " + summary + "\n Original abstract: " + abstract + "\n Your annotation: "}
        ]

        outputs = self.generator(messages, do_sample=True, temperature=0.01, max_new_tokens=self.max_tokens)
        generated_text = outputs[0]['generated_text'][-1]['content']

        pattern = r'annotation is:\s*(Yes|No)\b'
        match = re.search(pattern, generated_text, flags=re.IGNORECASE)
        if match:
            final_answer = match.group(1)
        else:
            match = re.search(r'(Yes|No)\s*$', generated_text, flags=re.IGNORECASE)
            if match:
                final_answer = match.group(1)
            else:
                print("empty")
                final_answer = 'Yes'

        return final_answer.lower().strip()



from typing import List, Dict, Set, Tuple

class Scorer(object):
    def keys(self) -> Set[str]:
        raise NotImplementedError

    def default_scores(self) -> Dict[str, float]:
        return {key: 0.0 for key in self.keys()}

    def score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> Tuple[Dict[str, float], List[Dict[str, float]]]:
        scores_dicts = self._score_single_ref(
            context,
            questions,
            answers,
            predictions,
            probabilities,
            null_probabilities
        )
        aggregated_scores = self.aggregate_scores(scores_dicts)
        return aggregated_scores, scores_dicts

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        raise NotImplementedError

    def score_multi_ref(
        self,
        context: str,
        questions_list: List[List[str]],
        answers_list: List[List[str]],
        predictions_list: List[List[str]],
        probabilities_list: List[List[float]],
        null_probabilities_list: List[List[float]]
    ) -> Tuple[Dict[str, float], List[List[Dict[str, float]]]]:
        reference_scores_list = []
        question_scores_list = []

        for i in range(len(questions_list)):
            reference_scores, question_scores = self.score_single_ref(
                context,
                questions_list[i],
                answers_list[i],
                predictions_list[i],
                probabilities_list[i],
                null_probabilities_list[i]
            )
            reference_scores_list.append(reference_scores)
            question_scores_list.append(question_scores)

        instance_scores = self.aggregate_scores(reference_scores_list)
        return instance_scores, question_scores_list

    def _ensure_expected_keys(self, expected_keys: Set[str], scores_dicts: List[Dict[str, float]]) -> None:
        for scores in scores_dicts:
            if expected_keys != scores.keys():
                raise Exception(f'Unequal keys: {expected_keys}; {scores.keys()}')

    def aggregate_scores(self, scores_dicts: List[Dict[str, float]]) -> Dict[str, float]:
        if len(scores_dicts) == 0:
            return self.default_scores()

        expected_keys = self.keys()
        self._ensure_expected_keys(expected_keys, scores_dicts)
        sums = {key: 0.0 for key in expected_keys}
        for scores in scores_dicts:
            for key in expected_keys:
                sums[key] += scores[key]

        averages = {key: sums[key] / len(scores_dicts) for key in expected_keys}
        return averages


from dataclasses import dataclass

@dataclass
class Config:
    classifier_type: str = 'learned'  # Options: 'learned', 'llama', 'gpt'
    classifier_path: str = '/path/to/classifier'
    question_generation_model_path: str = '/path/to/qa_generation'
    qa_answering_model_dir: str = '/path/to/qa_answering'
    llm_model_path: str = '/path/to/llm_model'


    input_file: str = '' # Supports .json, .csv, or .txt files
    delimiter: str = ','
    encoding: str = 'utf-8'

    ### ===Modify the heading/key names of your own PLS and abstracts=== ###
    target_sentence_col: str = 'Target_Sentence' # target summary/sentence
    abstract_col: str = 'Original_Abstract' # original abstract (source)
    input_file_format: str = 'csv'


    retrieval_k: int = 3 # Number of the retrieved snippets for each query
    knowledge_base: str = 'combined'  # Options: 'textbooks', 'statpearls', 'combined'

    answer_selection_strategy: str = 'llm-keywords'  # Options: 'llm-keywords', 'gpt-keywords', 'none'
    openai_api_key: str = 'your openai api key'

    cuda_device: int = 0
    verbose: bool = True
    generation_batch_size: int = 1
    answering_batch_size: int = 1
    scoring_batch_size: int = 1


from liquid import Template

general_cot_system = '''You are a helpful medical expert, and your task is to answer a multi-choice medical question. Please first think step-by-step and then choose the answer from the provided options. Organize your output in a json formatted as Dict{"step_by_step_thinking": Str(explanation), "answer_choice": Str{A/B/C/...}}. Your responses will be used for research purposes only, so please have a definite answer.'''

general_cot = Template('''
Here is the question:
{{question}}

Here are the potential choices:
{{options}}

Please think step-by-step and generate your output in json:
''')

general_medrag_system = '''You are a helpful medical expert, and your task is to answer a multi-choice medical question using the relevant documents. Please first think step-by-step and then choose the answer from the provided options. Organize your output in a json formatted as Dict{"step_by_step_thinking": Str(explanation), "answer_choice": Str{A/B/C/...}}. Your responses will be used for research purposes only, so please have a definite answer.'''

general_medrag = Template('''
Here are the relevant documents:
{{context}}

Here is the question:
{{question}}

Here are the potential choices:
{{options}}

Please think step-by-step and generate your output in json:
''')

meditron_cot = Template('''
### User:
Here is the question:
...

Here are the potential choices:
A. ...
B. ...
C. ...
D. ...
X. ...

Please think step-by-step and generate your output in json.

### Assistant:
{"step_by_step_thinking": ..., "answer_choice": "X"}

### User:
Here is the question:
{{question}}

Here are the potential choices:
{{options}}

Please think step-by-step and generate your output in json.

### Assistant:
''')

meditron_medrag = Template('''
Here are the relevant documents:
{{context}}

### User:
Here is the question:
...

Here are the potential choices:
A. ...
B. ...
C. ...
D. ...
X. ...

Please think step-by-step and generate your output in json.

### Assistant:
{"step_by_step_thinking": ..., "answer_choice": "X"}

### User:
Here is the question:
{{question}}

Here are the potential choices:
{{options}}

Please think step-by-step and generate your output in json.

### Assistant:
''')

simple_medrag_system = '''You are a helpful medical expert, and your task is to answer a medical question using the relevant documents.'''
simple_medrag_prompt = Template('''Here are the relevant documents:\n{{context}}\nHere is the question:\n{{question}}''')

i_medrag_system = '''You are a helpful medical assistant, and your task is to answer the given question following the instructions given by the user. '''

follow_up_instruction_ask = '''Please first analyze all the information in a section named Analysis (## Analysis). Then, use key terms from previous answers to form specific and direct questions. Generate {} concise, context-specific queries to search for additional information in an external knowledge base, in a section named Queries (## Queries). Each query should be simple and focused, directly relating to the key terms used in the answers. Wait for responses from the user before proceeding.'''
follow_up_instruction_answer = '''Please first think step-by-step to analyze all the information in a section named Analysis (## Analysis). Then, please provide your answer choice in a section named Answer (## Answer).'''


from qaeval.scoring.scorers.scorer import Scorer
from qaeval.scoring.scorers.is_answered import IsAnsweredScorer
from qaeval.scoring.scorers.em import ExactMatchScorer
from qaeval.scoring.scorers.f1 import F1Scorer
from qaeval.scoring.scorers.meta import MetaScorer


import math
import torch
import torch.nn.functional as F
from transformers import BartForConditionalGeneration, BartTokenizer
from tqdm import tqdm
from typing import Any, Dict, List, Tuple

class QuestionGenerationModel:
    def __init__(self,
                 model_path: str,
                 cuda_device: int = 0,
                 batch_size: int = 8,
                 silent: bool = True):
        """
        Initialize the question generation model
        
        Args:
            model_path (str): Path to the fine-tuned model
            cuda_device (int): GPU device number
            batch_size (int): Batch size for generation
            silent (bool): Whether to show progress bar
        """
        self.device = torch.device(f'cuda:{cuda_device}' if torch.cuda.is_available() else 'cpu')
        
        self.model = BartForConditionalGeneration.from_pretrained(model_path)
        self.tokenizer = BartTokenizer.from_pretrained(model_path)

        self.model.to(self.device)
        self.model.eval()
        
        self.batch_size = batch_size
        self.silent = silent

    def generate(self, text: str, start: int, end: int) -> str:
        """
        Generate a single question
        
        Args:
            text (str): Full context text
            start (int): Start index of answer
            end (int): End index of answer
        
        Returns:
            Generated question
        """
        return self.generate_all([(text, start, end)])[0]

    def generate_all(self, inputs: List[Tuple[str, int, int]]) -> List[str]:
        """
        Generate questions for multiple inputs
        
        Args:
            inputs (List[Tuple]): List of (text, start_index, end_index) tuples
        
        Returns:
            List of generated questions
        """
        outputs = []
        num_batches = int(math.ceil(len(inputs) / self.batch_size))
        generator = range(0, len(inputs), self.batch_size)
        if not self.silent:
            generator = tqdm(generator, total=num_batches, desc='Generating questions')

        for i in generator:
            batch = inputs[i:i + self.batch_size]
            batch_contexts = []
            batch_answers = []
            
            for text, start, end in batch:
                context = text
                answer = text[start:end]
                batch_contexts.append(context)
                batch_answers.append(answer)

            encodings = self.tokenizer(
                batch_contexts, 
                batch_answers, 
                padding=True, 
                truncation=True, 
                max_length=512, 
                return_tensors='pt'
            )

            input_ids = encodings['input_ids'].to(self.device)
            attention_mask = encodings['attention_mask'].to(self.device)

            with torch.no_grad():
                generated_ids = self.model.generate(
                    input_ids=input_ids,
                    attention_mask=attention_mask,
                    max_length=64,
                    num_return_sequences=1,
                    do_sample=False
                )

            batch_questions = self.tokenizer.batch_decode(
                generated_ids, 
                skip_special_tokens=True
            )
            
            outputs.extend(batch_questions)
        
        return outputs


import os
import tqdm
import json
import re
from langchain.text_splitter import RecursiveCharacterTextSplitter

def ends_with_ending_punctuation(s):
    ending_punctuation = ('.', '?', '!')
    return any(s.endswith(char) for char in ending_punctuation)

def concat(title, content):
    if ends_with_ending_punctuation(title.strip()):
        return title.strip() + " " + content.strip()
    else:
        return title.strip() + ". " + content.strip()

if __name__ == "__main__":

    text_splitter = RecursiveCharacterTextSplitter(chunk_size=1000, chunk_overlap=200)
    fdir = "corpus/textbooks/en"
    fnames = sorted(os.listdir(fdir))

    if not os.path.exists("corpus/textbooks/chunk"):
        os.makedirs("corpus/textbooks/chunk")
        
    for fname in tqdm.tqdm(fnames):
        fpath = os.path.join("corpus/textbooks/en", fname)
        texts = text_splitter.split_text(open(fpath).read().strip())
        saved_text = [json.dumps({"id": '_'.join([fname.replace(".txt", ''), str(i)]), "title": fname.strip(".txt"), "content": re.sub("\s+", " ", texts[i]), "contents": concat(fname.strip(".txt"), re.sub("\s+", " ", texts[i]))}) for i in range(len(texts))]
        with open("corpus/textbooks/chunk/{:s}".format(fname.replace(".txt", ".jsonl")), 'w') as f:
            f.write('\n'.join(saved_text))



from sentence_transformers.models import Transformer, Pooling
from sentence_transformers import SentenceTransformer
import os
import faiss
import json
import torch
import tqdm
import numpy as np

corpus_names = {
    "PubMed": ["pubmed"],
    "Textbooks": ["textbooks"],
    "StatPearls": ["statpearls"],
    "Wikipedia": ["wikipedia"],
    "MedText": ["textbooks", "statpearls"],
    "MedCorp": ["pubmed", "textbooks", "statpearls", "wikipedia"],
}

retriever_names = {
    "BM25": ["bm25"],
    "Contriever": ["facebook/contriever"],
    "SPECTER": ["allenai/specter"],
    "MedCPT": ["ncbi/MedCPT-Query-Encoder"],
    "RRF-2": ["bm25", "ncbi/MedCPT-Query-Encoder"],
    "RRF-4": ["bm25", "facebook/contriever", "allenai/specter", "ncbi/MedCPT-Query-Encoder"]
}

def ends_with_ending_punctuation(s):
    ending_punctuation = ('.', '?', '!')
    return any(s.endswith(char) for char in ending_punctuation)

def concat(title, content):
    if ends_with_ending_punctuation(title.strip()):
        return title.strip() + " " + content.strip()
    else:
        return title.strip() + ". " + content.strip()

class CustomizeSentenceTransformer(SentenceTransformer): # change the default pooling "MEAN" to "CLS"

    def _load_auto_model(self, model_name_or_path, *args, **kwargs):
        """
        Creates a simple Transformer + CLS Pooling model and returns the modules
        """
        print("No sentence-transformers model found with name {}. Creating a new one with CLS pooling.".format(model_name_or_path))
        token = kwargs.get('token', None)
        cache_folder = kwargs.get('cache_folder', None)
        revision = kwargs.get('revision', None)
        trust_remote_code = kwargs.get('trust_remote_code', False)
        if 'token' in kwargs or 'cache_folder' in kwargs or 'revision' in kwargs or 'trust_remote_code' in kwargs:
            transformer_model = Transformer(
                model_name_or_path,
                cache_dir=cache_folder,
                model_args={"token": token, "trust_remote_code": trust_remote_code, "revision": revision},
                tokenizer_args={"token": token, "trust_remote_code": trust_remote_code, "revision": revision},
            )
        else:
            transformer_model = Transformer(model_name_or_path)
        pooling_model = Pooling(transformer_model.get_word_embedding_dimension(), 'cls')
        return [transformer_model, pooling_model]


def embed(chunk_dir, index_dir, model_name, device, **kwarg):

    save_dir = os.path.join(index_dir, "embedding")
    
    if "contriever" in model_name:
        model = SentenceTransformer(model_name, device=device if torch.cuda.is_available() else "cpu")
    else:
        model = CustomizeSentenceTransformer(model_name, device=device if torch.cuda.is_available() else "cpu")

    model.eval()

    fnames = sorted([fname for fname in os.listdir(chunk_dir) if fname.endswith(".jsonl")])

    if not os.path.exists(save_dir):
        os.makedirs(save_dir)

    with torch.no_grad():
        for fname in tqdm.tqdm(fnames):
            fpath = os.path.join(chunk_dir, fname)
            save_path = os.path.join(save_dir, fname.replace(".jsonl", ".npy"))
            if os.path.exists(save_path):
                continue
            if open(fpath).read().strip() == "":
                continue
            texts = [json.loads(item) for item in open(fpath).read().strip().split('\n')]
            if "specter" in model_name.lower():
                texts = [model.tokenizer.sep_token.join([item["title"], item["content"]]) for item in texts]
            elif "contriever" in model_name.lower():
                texts = [". ".join([item["title"], item["content"]]).replace('..', '.').replace("?.", "?") for item in texts]
            elif "medcpt" in model_name.lower():
                texts = [[item["title"], item["content"]] for item in texts]
            else:
                texts = [concat(item["title"], item["content"]) for item in texts]
            embed_chunks = model.encode(texts, **kwarg)
            np.save(save_path, embed_chunks)
        embed_chunks = model.encode([""], **kwarg)
    return embed_chunks.shape[-1]

def construct_index(index_dir, model_name, h_dim=768, HNSW=False, M=32):

    with open(os.path.join(index_dir, "metadatas.jsonl"), 'w') as f:
        f.write("")
    
    if HNSW:
        M = M
        if "specter" in model_name.lower():
            index = faiss.IndexHNSWFlat(h_dim, M)
        else:
            index = faiss.IndexHNSWFlat(h_dim, M)
            index.metric_type = faiss.METRIC_INNER_PRODUCT
    else:
        if "specter" in model_name.lower():
            index = faiss.IndexFlatL2(h_dim)
        else:
            index = faiss.IndexFlatIP(h_dim)

    for fname in tqdm.tqdm(sorted(os.listdir(os.path.join(index_dir, "embedding")))):
        curr_embed = np.load(os.path.join(index_dir, "embedding", fname))
        index.add(curr_embed)
        with open(os.path.join(index_dir, "metadatas.jsonl"), 'a+') as f:
            f.write("\n".join([json.dumps({'index': i, 'source': fname.replace(".npy", "")}) for i in range(len(curr_embed))]) + '\n')

    faiss.write_index(index, os.path.join(index_dir, "faiss.index"))
    return index


class Retriever: 

    def __init__(self, retriever_name="ncbi/MedCPT-Query-Encoder", corpus_name="textbooks", db_dir="./corpus", HNSW=False, device='cuda:0', **kwarg):
        self.retriever_name = retriever_name
        self.corpus_name = corpus_name
        self.device = device
        self.db_dir = db_dir
        if not os.path.exists(self.db_dir):
            os.makedirs(self.db_dir)
        self.chunk_dir = os.path.join(self.db_dir, self.corpus_name, "chunk")
        if not os.path.exists(self.chunk_dir):
            print("Cloning the {:s} corpus from Huggingface...".format(self.corpus_name))
            os.system("git clone https://huggingface.co/datasets/MedRAG/{:s} {:s}".format(corpus_name, os.path.join(self.db_dir, self.corpus_name)))
            if self.corpus_name == "statpearls":
                print("Downloading the statpearls corpus from NCBI bookshelf...")
                os.system("wget https://ftp.ncbi.nlm.nih.gov/pub/litarch/3d/12/statpearls_NBK430685.tar.gz -P {:s}".format(os.path.join(self.db_dir, self.corpus_name)))
                os.system("tar -xzvf {:s} -C {:s}".format(os.path.join(db_dir, self.corpus_name, "statpearls_NBK430685.tar.gz"), os.path.join(self.db_dir, self.corpus_name)))
                print("Chunking the statpearls corpus...")
                os.system("python src/data/statpearls.py")
        self.index_dir = os.path.join(self.db_dir, self.corpus_name, "index", self.retriever_name.replace("Query-Encoder", "Article-Encoder"))
        if "bm25" in self.retriever_name.lower():
            from pyserini.search.lucene import LuceneSearcher
            self.metadatas = None
            self.embedding_function = None
            if os.path.exists(self.index_dir):
                self.index = LuceneSearcher(os.path.join(self.index_dir))
            else:
                os.system("python -m pyserini.index.lucene --collection JsonCollection --input {:s} --index {:s} --generator DefaultLuceneDocumentGenerator --threads 16".format(self.chunk_dir, self.index_dir))
                self.index = LuceneSearcher(os.path.join(self.index_dir))
        else:
            if os.path.exists(os.path.join(self.index_dir, "faiss.index")):
                self.index = faiss.read_index(os.path.join(self.index_dir, "faiss.index"))
                self.metadatas = [json.loads(line) for line in open(os.path.join(self.index_dir, "metadatas.jsonl")).read().strip().split('\n')]
            else:
                print("[In progress] Embedding the {:s} corpus with the {:s} retriever...".format(self.corpus_name, self.retriever_name.replace("Query-Encoder", "Article-Encoder")))
                if self.corpus_name in ["textbooks", "pubmed", "wikipedia"] and self.retriever_name in ["allenai/specter", "facebook/contriever", "ncbi/MedCPT-Query-Encoder"] and not os.path.exists(os.path.join(self.index_dir, "embedding")):
                    print("[In progress] Downloading the {:s} embeddings given by the {:s} model...".format(self.corpus_name, self.retriever_name.replace("Query-Encoder", "Article-Encoder")))
                    os.makedirs(self.index_dir, exist_ok=True)
                    if self.corpus_name == "textbooks":
                        if self.retriever_name == "allenai/specter":
                            os.system("wget -O {:s} https://myuva-my.sharepoint.com/:u:/g/personal/hhu4zu_virginia_edu/EYRRpJbNDyBOmfzCOqfQzrsBwUX0_UT8-j_geDPcVXFnig?download=1".format(os.path.join(self.index_dir, "embedding.zip")))
                        elif self.retriever_name == "facebook/contriever":
                            os.system("wget -O {:s} https://myuva-my.sharepoint.com/:u:/g/personal/hhu4zu_virginia_edu/EQqzldVMCCVIpiFV4goC7qEBSkl8kj5lQHtNq8DvHJdAfw?download=1".format(os.path.join(self.index_dir, "embedding.zip")))
                        elif self.retriever_name == "ncbi/MedCPT-Query-Encoder":
                            os.system("wget -O {:s} https://myuva-my.sharepoint.com/:u:/g/personal/hhu4zu_virginia_edu/EQ8uXe4RiqJJm0Tmnx7fUUkBKKvTwhu9AqecPA3ULUxUqQ?download=1".format(os.path.join(self.index_dir, "embedding.zip")))
                    elif self.corpus_name == "pubmed":
                        if self.retriever_name == "allenai/specter":
                            os.system("wget -O {:s} https://myuva-my.sharepoint.com/:u:/g/personal/hhu4zu_virginia_edu/Ebz8ySXt815FotxC1KkDbuABNycudBCoirTWkKfl8SEswA?download=1".format(os.path.join(self.index_dir, "embedding.zip")))
                        elif self.retriever_name == "facebook/contriever":
                            os.system("wget -O {:s} https://myuva-my.sharepoint.com/:u:/g/personal/hhu4zu_virginia_edu/EWecRNfTxbRMnM0ByGMdiAsBJbGJOX_bpnUoyXY9Bj4_jQ?download=1".format(os.path.join(self.index_dir, "embedding.zip")))
                        elif self.retriever_name == "ncbi/MedCPT-Query-Encoder":
                            os.system("wget -O {:s} https://myuva-my.sharepoint.com/:u:/g/personal/hhu4zu_virginia_edu/EVCuryzOqy5Am5xzRu6KJz4B6dho7Tv7OuTeHSh3zyrOAw?download=1".format(os.path.join(self.index_dir, "embedding.zip")))
                    elif self.corpus_name == "wikipedia":
                        if self.retriever_name == "allenai/specter":
                            os.system("wget -O {:s} https://myuva-my.sharepoint.com/:u:/g/personal/hhu4zu_virginia_edu/Ed7zG3_ce-JOmGTbgof3IK0BdD40XcuZ7AGZRcV_5D2jkA?download=1".format(os.path.join(self.index_dir, "embedding.zip")))
                        elif self.retriever_name == "facebook/contriever":
                            os.system("wget -O {:s} https://myuva-my.sharepoint.com/:u:/g/personal/hhu4zu_virginia_edu/ETKHGV9_KNBPmDM60MWjEdsBXR4P4c7zZk1HLLc0KVaTJw?download=1".format(os.path.join(self.index_dir, "embedding.zip")))
                        elif self.retriever_name == "ncbi/MedCPT-Query-Encoder":
                            os.system("wget -O {:s} https://myuva-my.sharepoint.com/:u:/g/personal/hhu4zu_virginia_edu/EXoxEANb_xBFm6fa2VLRmAcBIfCuTL-5VH6vl4GxJ06oCQ?download=1".format(os.path.join(self.index_dir, "embedding.zip")))
                    os.system("unzip {:s} -d {:s}".format(os.path.join(self.index_dir, "embedding.zip"), self.index_dir))
                    os.system("rm {:s}".format(os.path.join(self.index_dir, "embedding.zip")))
                    h_dim = 768
                else:
                    h_dim = embed(chunk_dir=self.chunk_dir, index_dir=self.index_dir, model_name=self.retriever_name.replace("Query-Encoder", "Article-Encoder"), device=self.device, **kwarg)

                print("[In progress] Embedding finished! The dimension of the embeddings is {:d}.".format(h_dim))
                self.index = construct_index(index_dir=self.index_dir, model_name=self.retriever_name.replace("Query-Encoder", "Article-Encoder"), h_dim=h_dim, HNSW=HNSW)
                print("[Finished] Corpus indexing finished!")
                self.metadatas = [json.loads(line) for line in open(os.path.join(self.index_dir, "metadatas.jsonl")).read().strip().split('\n')]            
            if "contriever" in self.retriever_name.lower():
                self.embedding_function = SentenceTransformer(self.retriever_name, device=device if torch.cuda.is_available() else "cpu")
            else:
                self.embedding_function = CustomizeSentenceTransformer(self.retriever_name, device=device if torch.cuda.is_available() else "cpu")
            self.embedding_function.eval()

    def get_relevant_documents(self, question, k=32, id_only=False, **kwarg):
        assert type(question) == str
        question = [question]

        if "bm25" in self.retriever_name.lower():
            res_ = [[]]
            hits = self.index.search(question[0], k=k)
            res_[0].append(np.array([h.score for h in hits]))
            ids = [h.docid for h in hits]
            indices = [{"source": '_'.join(h.docid.split('_')[:-1]), "index": eval(h.docid.split('_')[-1])} for h in hits]
        else:
            with torch.no_grad():
                query_embed = self.embedding_function.encode(question, **kwarg)
            res_ = self.index.search(query_embed, k=k)
            ids = ['_'.join([self.metadatas[i]["source"], str(self.metadatas[i]["index"])]) for i in res_[1][0]]
            indices = [self.metadatas[i] for i in res_[1][0]]

        scores = res_[0][0].tolist()
        
        if id_only:
            return [{"id":i} for i in ids], scores
        else:
            return self.idx2txt(indices), scores

    def idx2txt(self, indices): # return List of Dict of str
        '''
        Input: List of Dict( {"source": str, "index": int} )
        Output: List of str
        '''
        return [json.loads(open(os.path.join(self.chunk_dir, i["source"]+".jsonl")).read().strip().split('\n')[i["index"]]) for i in indices]

class RetrievalSystem:

    def __init__(self, retriever_name="MedCPT", corpus_name="Textbooks", db_dir="./corpus", HNSW=False, cache=False, device='cuda:0'):
        self.retriever_name = retriever_name
        self.corpus_name = corpus_name
        self.device = device
        assert self.corpus_name in corpus_names
        assert self.retriever_name in retriever_names
        self.retrievers = []
        for retriever in retriever_names[self.retriever_name]:
            self.retrievers.append([])
            for corpus in corpus_names[self.corpus_name]:
                self.retrievers[-1].append(Retriever(retriever, corpus, db_dir, HNSW=HNSW, device=self.device))
        self.cache = cache
        if self.cache:
            self.docExt = DocExtracter(cache=True, corpus_name=self.corpus_name, db_dir=db_dir)
        else:
            self.docExt = None
    
    def retrieve(self, question, k=32, rrf_k=100, id_only=False):
        '''
            Given questions, return the relevant snippets from the corpus
        '''
        assert type(question) == str

        output_id_only = id_only
        if self.cache:
            id_only = True

        texts = []
        scores = []

        if "RRF" in self.retriever_name:
            k_ = max(k * 2, 100)
        else:
            k_ = k
        for i in range(len(retriever_names[self.retriever_name])):
            texts.append([])
            scores.append([])
            for j in range(len(corpus_names[self.corpus_name])):
                t, s = self.retrievers[i][j].get_relevant_documents(question, k=k_, id_only=id_only)
                texts[-1].append(t)
                scores[-1].append(s)
        texts, scores = self.merge(texts, scores, k=k, rrf_k=rrf_k)
        if self.cache:
            texts = self.docExt.extract(texts)
        return texts, scores

    def merge(self, texts, scores, k=32, rrf_k=100):
        '''
            Merge the texts and scores from different retrievers
        '''
        RRF_dict = {}
        for i in range(len(retriever_names[self.retriever_name])):
            texts_all, scores_all = None, None
            for j in range(len(corpus_names[self.corpus_name])):
                if texts_all is None:
                    texts_all = texts[i][j]
                    scores_all = scores[i][j]
                else:
                    texts_all = texts_all + texts[i][j]
                    scores_all = scores_all + scores[i][j]
            if "specter" in retriever_names[self.retriever_name][i].lower():
                sorted_index = np.array(scores_all).argsort()
            else:
                sorted_index = np.array(scores_all).argsort()[::-1]
            texts[i] = [texts_all[i] for i in sorted_index]
            scores[i] = [scores_all[i] for i in sorted_index]
            for j, item in enumerate(texts[i]):
                if item["id"] in RRF_dict:
                    RRF_dict[item["id"]]["score"] += 1 / (rrf_k + j + 1)
                    RRF_dict[item["id"]]["count"] += 1
                else:
                    RRF_dict[item["id"]] = {
                        "id": item["id"],
                        "title": item.get("title", ""),
                        "content": item.get("content", ""),
                        "score": 1 / (rrf_k + j + 1),
                        "count": 1
                        }
        RRF_list = sorted(RRF_dict.items(), key=lambda x: x[1]["score"], reverse=True)
        if len(texts) == 1:
            texts = texts[0][:k]
            scores = scores[0][:k]
        else:
            texts = [dict((key, item[1][key]) for key in ("id", "title", "content")) for item in RRF_list[:k]]
            scores = [item[1]["score"] for item in RRF_list[:k]]
        return texts, scores
    

class DocExtracter:
    
    def __init__(self, db_dir="./corpus", cache=False, corpus_name="MedCorp"):
        self.db_dir = db_dir
        self.cache = cache
        print("Initializing the document extracter...")
        for corpus in corpus_names[corpus_name]:
            if not os.path.exists(os.path.join(self.db_dir, corpus, "chunk")):
                print("Cloning the {:s} corpus from Huggingface...".format(corpus))
                os.system("git clone https://huggingface.co/datasets/MedRAG/{:s} {:s}".format(corpus, os.path.join(self.db_dir, corpus)))
                if corpus == "statpearls":
                    print("Downloading the statpearls corpus from NCBI bookshelf...")
                    os.system("wget https://ftp.ncbi.nlm.nih.gov/pub/litarch/3d/12/statpearls_NBK430685.tar.gz -P {:s}".format(os.path.join(self.db_dir, corpus)))
                    os.system("tar -xzvf {:s} -C {:s}".format(os.path.join(self.db_dir, corpus, "statpearls_NBK430685.tar.gz"), os.path.join(self.db_dir, corpus)))
                    print("Chunking the statpearls corpus...")
                    os.system("python src/data/statpearls.py")
        if self.cache:
            if os.path.exists(os.path.join(self.db_dir, "_".join([corpus_name, "id2text.json"]))):
                self.dict = json.load(open(os.path.join(self.db_dir, "_".join([corpus_name, "id2text.json"]))))
            else:
                self.dict = {}
                for corpus in corpus_names[corpus_name]:
                    for fname in tqdm.tqdm(sorted(os.listdir(os.path.join(self.db_dir, corpus, "chunk")))):
                        if open(os.path.join(self.db_dir, corpus, "chunk", fname)).read().strip() == "":
                            continue
                        for i, line in enumerate(open(os.path.join(self.db_dir, corpus, "chunk", fname)).read().strip().split('\n')):
                            item = json.loads(line)
                            _ = item.pop("contents", None)
                            # assert item["id"] not in self.dict
                            self.dict[item["id"]] = item
                with open(os.path.join(self.db_dir, "_".join([corpus_name, "id2text.json"])), 'w') as f:
                    json.dump(self.dict, f)
        else:
            if os.path.exists(os.path.join(self.db_dir, "_".join([corpus_name, "id2path.json"]))):
                self.dict = json.load(open(os.path.join(self.db_dir, "_".join([corpus_name, "id2path.json"]))))
            else:
                self.dict = {}
                for corpus in corpus_names[corpus_name]:
                    for fname in tqdm.tqdm(sorted(os.listdir(os.path.join(self.db_dir, corpus, "chunk")))):
                        if open(os.path.join(self.db_dir, corpus, "chunk", fname)).read().strip() == "":
                            continue
                        for i, line in enumerate(open(os.path.join(self.db_dir, corpus, "chunk", fname)).read().strip().split('\n')):
                            item = json.loads(line)
                            # assert item["id"] not in self.dict
                            self.dict[item["id"]] = {"fpath": os.path.join(corpus, "chunk", fname), "index": i}
                with open(os.path.join(self.db_dir, "_".join([corpus_name, "id2path.json"])), 'w') as f:
                    json.dump(self.dict, f, indent=4)
        print("Initialization finished!")
    
    def extract(self, ids):
        if self.cache:
            output = []
            for i in ids:
                item = self.dict[i] if type(i) == str else self.dict[i["id"]]
                output.append(item)
        else:
            output = []
            for i in ids:
                item = self.dict[i] if type(i) == str else self.dict[i["id"]]
                output.append(json.loads(open(os.path.join(self.db_dir, item["fpath"])).read().strip().split('\n')[item["index"]]))
        return output



import argparse
import sys
import os
import pandas as pd
import numpy as np
from transformers import pipeline
from src.medrag import MedRAG
from plainqafact import PlainQAFact
from classifier import LearnedClassifier, LLMClassifier
from default_config import Config
from openai import OpenAI


def parse_args():
    parser = argparse.ArgumentParser(description='Run medical text evaluation pipeline')
    parser.add_argument('--classifier_path', help='Path to input CSV file')
    parser.add_argument('--input_file', required=True, help='Path to input file (.csv, .json, or .txt)')
    parser.add_argument('--input_file_format', choices=['csv', 'json', 'txt'], default='csv',
                        help='Input file format')
    parser.add_argument('--delimiter', default=',', help='Delimiter for csv/txt files')
    parser.add_argument('--encoding', default='utf-8', help='File encoding (default: utf-8)')
    parser.add_argument('--question_generation_model_path', required=True, help='Path to QA generation model')
    parser.add_argument('--qa_answering_model_dir', required=True, help='Path to QA answering model directory')
    parser.add_argument('--llm_model_path', required=True, help='Path to the Llama model')
    parser.add_argument('--cuda_device', default='cuda:0', help='Define your cuda device (e.g., cuda:0)')
    parser.add_argument('--scoring_batch_size', default=1, help='Batch size of BERTScore evaluation')
    parser.add_argument('--verbose', default=True)
    parser.add_argument(
        '--classifier_type',
        choices=['learned', 'llama', 'gpt'],
        default='learned',
        help='Type of classifier to use (default: learned)'
    )
    parser.add_argument(
        '--knowledge_base',
        choices=['textbooks', 'statpearls', 'combined'],
        default='combined',
        help='Knowledge base to use for retrieval (default: combined)'
    )
    parser.add_argument(
        '--answer_selection_strategy',
        choices=['llm-keywords', 'gpt-keywords', 'none'],
        default='llm-keywords',
        help='Strategy for answer selection (default: llm-keywords)'
    )
    return parser.parse_args()

def initialize_answer_extractor(config: Config):
    if config.answer_selection_strategy == 'llm-keywords':
        return pipeline(
            "text-generation",
            model=config.llm_model_path,
            device=config.cuda_device
        )
    elif config.answer_selection_strategy == 'gpt-keywords':
        return OpenAI(api_key=config.openai_api_key)
    else:
        return None

def initialize_classifier(config: Config):
    if config.classifier_type == 'learned':
        return LearnedClassifier(config.classifier_path, device=config.cuda_device)
    elif config.classifier_type == 'llama':
        return LLMClassifier('llama', model_path=config.classifier_path, device=config.cuda_device)
    elif config.classifier_type == 'gpt':
        return LLMClassifier('gpt', openai_key=config.openai_api_key)
    else:
        raise ValueError(f"Unknown classifier type: {config.classifier_type}")

def process_data(config: Config):
    if config.input_file_format == 'csv':
        df = pd.read_csv(config.input_file, delimiter=config.delimiter, encoding=config.encoding)
    elif config.input_file_format == 'json':
        df = pd.read_json(config.input_file, encoding=config.encoding)
    elif config.input_file_format == 'txt':
        df = pd.read_csv(config.input_file, delimiter=config.delimiter, encoding=config.encoding)
    else:
        raise ValueError(f"Unsupported file format: {config.input_file_format}")

    if config.target_sentence_col not in df.columns or config.abstract_col not in df.columns:
        raise ValueError(f"Required columns '{config.target_sentence_col}' and '{config.abstract_col}' "
                        f"must be present in the input file")

    sentences = df[config.target_sentence_col]
    abstracts = df[config.abstract_col]

    classifier = initialize_classifier(config)
    external, external_abs, internal, internal_abs = [], [], [], []
    for abstract, sentence in zip(abstracts, sentences):
        label, _ = classifier.predict(abstract, sentence)
        if label == 'yes':
            external.append(sentence)
            external_abs.append(abstract)
        else:
            internal.append(sentence)
            internal_abs.append(abstract)

    return prepare_evaluation_data(external, external_abs, internal, internal_abs)

def prepare_evaluation_data(external, external_abs, internal, internal_abs):
    external_pls = [[s] for s in external]
    internal_pls = [[s] for s in internal]
    return {
        'external': {'abstracts': external_abs, 'summaries': external_pls},
        'internal': {'abstracts': internal_abs, 'summaries': internal_pls}
    }

def retrieve_knowledge(queries, config: Config, corpus_name: str):
    knowledge = []
    for query in queries:
        medrag = MedRAG(retrieval=True, retriever_name="MedCPT", corpus_name=corpus_name, corpus_cache=True)
        snippets, _ = medrag.answer(question=query[0], k=config.retrieval_k)
        content = ' '.join([item['content'].strip().replace('\n', '') for item in snippets])
        knowledge.append(content)
    return knowledge

def get_combined_knowledge(abstracts, summaries, config: Config):
    if config.knowledge_base == 'textbooks':
        knowledge = retrieve_knowledge(summaries, config, "Textbooks")
        return [f"{abs_text}{tb_know}"
                for abs_text, tb_know in zip(abstracts, knowledge)]

    elif config.knowledge_base == 'statpearls':
        knowledge = retrieve_knowledge(summaries, config, "StatPearls")
        return [f"{abs_text}{sp_know}"
                for abs_text, sp_know in zip(abstracts, knowledge)]

    else:
        textbook_knowledge = retrieve_knowledge(summaries, config, "Textbooks")
        statpearl_knowledge = retrieve_knowledge(summaries, config, "StatPearls")
        return [f"{abs_text} {tb_know} {sp_know}"
                for abs_text, tb_know, sp_know in zip(abstracts, textbook_knowledge, statpearl_knowledge)]

def main():
    args = parse_args()
    config = Config()
    for key, value in vars(args).items():
        setattr(config, key, value)

    data = process_data(config)

    combined_contexts = get_combined_knowledge(
        data['external']['abstracts'],
        data['external']['summaries'],
        config
    )

    metric = PlainQAFact(
        classifier_type=config.classifier_type,
        classifier_path=config.classifier_path,
        generation_model_path=config.question_generation_model_path,
        answering_model_dir=config.qa_answering_model_dir,
        cuda_device=config.cuda_device,
        knowledge_base=config.knowledge_base,
        generation_batch_size=config.generation_batch_size,
        answering_batch_size=config.answering_batch_size,
        scoring_batch_size=config.scoring_batch_size,
        answer_selection_strategy=config.answer_selection_strategy
    )

    try:
        generator = initialize_answer_extractor(config)
    except Exception as e:
        print(f"Error initializing generator: {str(e)}")
        sys.exit(1)

    # Run evaluation
    external_results = metric.score_batch_qafacteval(
        combined_contexts,
        data['external']['summaries'],
        return_qa_pairs=True,
        generator=generator
    )

    internal_results = metric.score_batch_qafacteval(
        data['internal']['abstracts'],
        data['internal']['summaries'],
        return_qa_pairs=True,
        generator=generator
    )

    external_scores = [metrics['qa-eval']['bertscore'] for metrics, *_ in external_results]
    internal_scores = [metrics['qa-eval']['bertscore'] for metrics, *_ in internal_results]

    print(f"Explanation scores (mean: {np.mean(external_scores):.4f}):", external_scores)
    print(f"Simplification scores (mean: {np.mean(internal_scores):.4f}):", internal_scores)
    print(f"PlainQAFact score: {np.mean(external_scores + internal_scores):.4f}")


if __name__ == "__main__":
    main()


import torch
from torch.utils.data import DataLoader, SequentialSampler
from tqdm import tqdm
from transformers import (
    AutoConfig,
    AutoModelForQuestionAnswering,
    pipeline,
    squad_convert_examples_to_features,
)
from transformers.data.processors.squad import SquadResult, SquadExample
from typing import Dict, List, Tuple, Union
from qaeval.answering.utils import compute_predictions_logits_with_null, fix_answer_span, SpanFixError

Prediction = Union[
    Tuple[str, float, float],
    Tuple[str, float, float, Tuple[int, int]],
    Dict[str, Union[str, float, Tuple[int, int]]],
]
import math
from transformers_old_tokenizer.tokenization_auto import AutoTokenizer

class QuestionAnsweringModel(object):
    def __init__(self,
                 model_dir: str,
                 cuda_device: int = 0,
                 batch_size: int = 8,
                 silent: bool = True) -> None:
        self.config = AutoConfig.from_pretrained(model_dir)
        self.tokenizer = AutoTokenizer.from_pretrained(model_dir)
        self.model = AutoModelForQuestionAnswering.from_pretrained(model_dir, config=self.config)
        self.device = torch.device(f"cuda:{cuda_device}" if torch.cuda.is_available() else "cpu")
        if self.device != 'cpu':
            self.model.to(self.device)

        self.model_type = 'electra'
        self.batch_size = batch_size
        self.max_seq_length = 512 #384
        self.doc_stride = 128
        self.silent = silent

    def _to_list(self, tensor):
        return tensor.detach().cpu().tolist()

    def _try_fixing_offsets(
        self,
        contexts: List[str],
        predictions: Dict[str, str],
        offsets_dict: Dict[str, Tuple[int, int]],
    ) -> Dict[str, Tuple[int, int]]:
        """
        Tries to fix the potentially noisy character offsets of the predictions in the `contexts`.
        The input and output end indices are exclusive.
        """
        new_offsets = {}

        for i, context in enumerate(contexts):
            index = str(i)
            if predictions[index]:
                prediction = predictions[index]
                pred_start, pred_end = offsets_dict[index]
            else: 
                prediction = None
                pred_start, pred_end = None, None
            if context is None or prediction is None or pred_start is None or pred_end is None:
                new_offsets[index] = (pred_start, pred_end)
            else:
                span = context[pred_start:pred_end]
                if span != prediction:
                    try:
                        pred_start, pred_end = fix_answer_span(prediction, span, pred_start, pred_end)
                    except SpanFixError:
                        pass
                new_offsets[index] = (pred_start, pred_end)
        return new_offsets

    def answer(
        self,
        question: str,
        context: str,
        return_offsets: bool = False,
        try_fixing_offsets: bool = True,
        return_dict: bool = False,
    ) -> Prediction:
        """
        Returns a tuple of (prediction, probability, null_probability). If `return_offsets = True`, the tuple
        will include rough character offsets of where the prediction is in the context. Because the tokenizer that
        the QA model uses does not support returning the character offsets from the BERT tokenization, we cannot
        directly provide exactly where the answer came from. However, the offsets should be pretty close to the
        prediction, and the prediction should be a substring of the offsets (modulo whitespace). If
        `return_offsets` and `try_fixing_offsets` are `True`, we will try to fix the character offsets via
        an alignment. See below.

        The `SquadExample` class maintains a list of whitespace separated tokens `doc_tokens` and a mapping
        from the context string characters to the token indices `char_to_word_offset`. Whitespace
        is included in the previous token. The `squad_convert_example_to_features` method takes each of these
        tokens and breaks it into the subtokens with the transformers tokenizer, which are passed into the model.
        It also keeps a mapping from the subtokens to the `doc_tokens` called `tok_to_orig_index`. The QA model
        predicts a span in the subtokens. In the `_get_char_offsets` method, we use these data structures to map
        from the subtoken span to character offsets. However, we cannot separate subtokens, so they are merged together.
        See the below example

            context: " My name is  Dan!"
            doc_tokens: [My, name, is, Dan!]
            char_to_word_offset: [-1, 0, 0, 0, 1, 1, 1, 1, 1, 2, 2, 2, 2, 3, 3, 3, 3]
            subtokens: [My, name, is, Dan, ##!]
            tok_to_orig_index: [0, 1, 2, 3, 3]

            prediction: "name is Dan"
            prediction subtokens: [name, is, Dan]
            prediction in doc_tokens: [name, is, Dan!]
            prediction in context: "name is  Dan!"

        The prediction includes the extra whitespace between "is" and "Dan" as well as the "!"

        If `try_fixing_offsets=True`, we will try to fix the character offsets to be correct based on an alignment
        algorithm. We use the `edlib` python package to create a character alignment between the actual prediction
        string and the span given by the original offsets. We then update the offsets based on the alignment. If
        this procedure fails, the original offsets will be returned.
        """
        return self.answer_all(
            [(question, context)], return_offsets=return_offsets,
            try_fixing_offsets=try_fixing_offsets, return_dicts=return_dict
        )[0]

    def answer_all(
        self,
        input_data: List[Tuple[str, str]],
        return_offsets: bool = False,
        try_fixing_offsets: bool = True,
        return_dicts: bool = False,
    ) -> List[Prediction]:
        examples = []
        for i, (question, context) in enumerate(input_data):
            if not question or not context:
                print(f"Question or context is empty for example {i}")
                print(f"Question: {question}")
                print(f"Context: {context}")
                question = context or "empty question"
                context = context or "empty context"
            question = str(question)
            context = str(context)
            examples.append(SquadExample(
                qas_id=str(i),
                question_text=question,
                context_text=context,
                answer_text=None,
                start_position_character=None,
                title=None,
                is_impossible=True,
                answers=[]
            ))

        features, dataset = squad_convert_examples_to_features(
            examples=examples,
            tokenizer=self.tokenizer,
            max_seq_length=self.max_seq_length,
            doc_stride=self.doc_stride,
            max_query_length=64, 
            is_training=False,
            return_dataset="pt",
            threads=1,
            tqdm_enabled=not self.silent
        )

        eval_sampler = SequentialSampler(dataset)
        eval_dataloader = DataLoader(dataset, sampler=eval_sampler, batch_size=self.batch_size)

        self.model.eval()
        all_results = []
        generator = eval_dataloader
        if not self.silent:
            generator = tqdm(generator, desc='Evaluating')

        for batch in generator:
            if self.device != 'cpu':
                batch = tuple(t.to(self.device) for t in batch)

            with torch.no_grad():
                inputs = {
                    "input_ids": batch[0],
                    "attention_mask": batch[1],
                    "token_type_ids": batch[2],
                }

                feature_indices = batch[3]
                outputs = self.model(**inputs)

            for i, feature_index in enumerate(feature_indices):
                eval_feature = features[feature_index.item()]
                unique_id = int(eval_feature.unique_id)

                if isinstance(outputs, dict):
                    start_logits = self._to_list(outputs['start_logits'][i])
                    end_logits = self._to_list(outputs['end_logits'][i])
                else:
                    start_logits = self._to_list(outputs[0][i])
                    end_logits = self._to_list(outputs[1][i])
                
                result = SquadResult(unique_id, start_logits, end_logits)
                all_results.append(result)

        model_predictions = compute_predictions_logits_with_null(
            self.tokenizer,
            examples,
            features,
            all_results,
            20,
            30,
            True,
            False,
            True,
            return_offsets=return_offsets
        )

        if return_offsets:
            predictions, prediction_probs, no_answer_probs, offsets = model_predictions
            if try_fixing_offsets:
                contexts = [context for _, context in input_data]
                offsets = self._try_fixing_offsets(contexts, predictions, offsets)
        else:
            predictions, prediction_probs, no_answer_probs = model_predictions

        results = []
        for i in range(len(input_data)):
            i = str(i)
            r = (predictions[i], prediction_probs[i], no_answer_probs[i])
            if return_dicts:
                r = {
                    'prediction': r[0],
                    'probability': r[1],
                    'null_probability': r[2],
                }

            if return_offsets:
                if return_dicts:
                    r['start'] = offsets[i][0]
                    r['end'] = offsets[i][1]
                else:
                    r = r + (offsets[i],)
            results.append(r)
        return results


from typing import Dict, List, Set
import bert_score
import numpy as np
from qaeval.scoring.scorers.scorer import Scorer

class BertScoreScorer(Scorer):
    def __init__(self, cuda_device: int, batch_size: int = 8) -> None:
        self.device = f'cuda:{cuda_device}' if cuda_device >= 0 else 'cpu'
        self.batch_size = batch_size

    def keys(self) -> Set[str]:
        return {'bertscore'}

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        indices = []
        valid_answers = []
        valid_predictions = []
        
        for i, (answer, prediction, probability, null_probability) in \
                enumerate(zip(answers, predictions, probabilities, null_probabilities)):
            if probability > null_probability:
                valid_answers.append(answer)
                valid_predictions.append(prediction)
                indices.append(i)

        scores = [0.0] * len(questions)
        if valid_predictions:
            _, _, F1 = bert_score.score(valid_predictions, valid_answers, 
                                      lang="en", device=self.device)
            for idx, score in zip(indices, F1.tolist()):
                scores[idx] = score

        return [{'bertscore': s} for s in scores]


from typing import Dict, List, Union
import sys
from transformers.data.metrics.squad_metrics import compute_f1
from qaeval.metric import QAEval
import bert_score
from transformers import pipeline
from bertscore_scorer import BertScoreScorer
import pandas as pd
import json
import numpy as np
from classifier import LearnedClassifier, LLMClassifier
from openai import OpenAI
from src.medrag import MedRAG
import torch
import gc

MetricsDict = Dict[str, float]
SummaryType = Union[str, List[str]]

def get_filter(qa_summ, answers_ref):
    answerable = []
    a_orig = []
    for qa_summ_ in qa_summ:
        answerability = (qa_summ_[1] > qa_summ_[2])
        answerable.append(answerability)
        a_orig.append(qa_summ_[0])

    f1s = [compute_f1(answer, prediction) for \
        answer, prediction in zip(answers_ref, a_orig)]
    bool_f1 = [x > 0.60 for x in f1s]
    bool_total = [x and y for x, y in zip(bool_f1, answerable)]
    return bool_total

class PlainQAFact(QAEval):
    def __init__(
            self,
            cuda_device: int,
            scoring_batch_size: int,
            answer_selection_strategy: str,
            openai_api_key: str = '',
            classifier_type: str = None,
            classifier_path: str = None,
            llm_model_path: str = None,
            question_generation_model_path: str = None,
            qa_answering_model_dir: str = None,
            knowledge_base: str = None,
            target_sentence_col: str = 'Target_Sentence',
            abstract_col: str = 'Original_Abstract',
            input_file_format: str = 'csv',
            delimiter: str = ',',
            encoding: str = 'utf-8',
            retrieval_k: int = 3,
            generation_batch_size: int = 1,
            answering_batch_size: int = 1,
            verbose: bool = True,
            *args,
            **kwargs):

        import os
        os.environ['PYTORCH_CUDA_ALLOC_CONF'] = 'expandable_segments:True'
        
        self.cuda_device = cuda_device
        self.classifier_type = classifier_type
        self.classifier_path = classifier_path
        self.llm_model_path = llm_model_path
        self.question_generation_model_path = question_generation_model_path
        self.qa_answering_model_dir = qa_answering_model_dir
        self.knowledge_base = knowledge_base
        self.answer_selection_strategy = answer_selection_strategy
        self.openai_api_key = openai_api_key
        self.target_sentence_col = target_sentence_col
        self.abstract_col = abstract_col
        self.input_file_format = input_file_format
        self.delimiter = delimiter
        self.encoding = encoding
        self.retrieval_k = retrieval_k
        self.generation_batch_size = generation_batch_size
        self.answering_batch_size = answering_batch_size
        self.scoring_batch_size = scoring_batch_size
            
        try:
            import spacy
            spacy.load("en_core_web_sm")
        except:
            import spacy.cli
            spacy.cli.download("en_core_web_sm")

        qaeval_params = {
            'cuda_device': cuda_device,
            'generation_model_path': question_generation_model_path,
            'answering_model_dir': qa_answering_model_dir, 
            'generation_batch_size': generation_batch_size,
            'answering_batch_size': answering_batch_size, 
            'verbose': verbose
        }
        
        plainqafact_params = [
            'classifier_type', 'classifier_path', 'llm_model_path',
            'question_generation_model_path', 'qa_answering_model_dir',
            'knowledge_base', 'target_sentence_col', 'abstract_col',
            'input_file_format', 'delimiter', 'encoding', 'retrieval_k',
            'generation_batch_size', 'answering_batch_size', 'openai_api_key',
            'scoring_batch_size', 'answer_selection_strategy'
        ]
        
        for key, value in kwargs.items():
            if key not in plainqafact_params:
                qaeval_params[key] = value

        self._answer_extractor = None
        self._classifier = None
        self._qa_model = None
        self._bertscore_scorer = None

        self._prevent_qa_model_load = True
        super().__init__(**qaeval_params)
        self._prevent_qa_model_load = False

    def __del__(self):
        try:
            if hasattr(self, '_classifier') and self._classifier is not None:
                del self._classifier
            if hasattr(self, '_qa_model') and self._qa_model is not None:
                del self._qa_model
            if hasattr(self, '_answer_extractor') and self._answer_extractor is not None:
                del self._answer_extractor
            if hasattr(self, '_bertscore_scorer') and self._bertscore_scorer is not None:
                del self._bertscore_scorer
            
            if torch.cuda.is_available():
                torch.cuda.empty_cache()
                gc.collect()
        except Exception as e:
            print(f"Warning: Error during cleanup: {str(e)}")

    def _initialize_answer_extractor(self):
        if self._answer_extractor is None:
            try:
                if self.answer_selection_strategy == 'llm-keywords':
                    model_kwargs = {
                        "device_map": "auto",
                        "torch_dtype": torch.float16,
                        "low_cpu_mem_usage": True
                    }
                    self._answer_extractor = pipeline(
                        "text-generation",
                        model=self.llm_model_path,
                        model_kwargs=model_kwargs
                    )
                elif self.answer_selection_strategy == 'gpt-keywords':
                    self._answer_extractor = OpenAI(api_key=self.openai_api_key)
                else:
                    self._answer_extractor = None
            except Exception as e:
                print(f"Warning: Failed to initialize answer extractor: {str(e)}")
                self._answer_extractor = None
                if torch.cuda.is_available():
                    torch.cuda.empty_cache()
                    gc.collect()
                
        return self._answer_extractor

    def _initialize_bertscore_scorer(self):
        if self._bertscore_scorer is None:
            self._bertscore_scorer = BertScoreScorer(cuda_device=self.cuda_device, batch_size=self.scoring_batch_size)
            self.scorer.scorers.append(self._bertscore_scorer)
        return self._bertscore_scorer

    def _initialize_classifier(self):
        if self._classifier is None:
            if self.classifier_type == 'learned':
                self._classifier = LearnedClassifier(self.classifier_path, device=self.cuda_device if torch.cuda.is_available() else "cpu")
            elif self.classifier_type == 'llama':
                self._classifier = LLMClassifier('llama', model_path=self.classifier_path, device=self.cuda_device if torch.cuda.is_available() else "cpu")
            elif self.classifier_type == 'gpt':
                self._classifier = LLMClassifier('gpt', openai_key=self.openai_api_key)
            else:
                raise ValueError(f"Unknown classifier type: {self.classifier_type}")
        return self._classifier

    def _retrieve_knowledge(self, queries, corpus_name: str):
        knowledge = []
        for query in queries:
            medrag = MedRAG(cuda_device=self.cuda_device, retrieval=True, retriever_name="MedCPT", corpus_name=corpus_name, corpus_cache=True)
            snippets, _ = medrag.answer(question=query[0], k=self.retrieval_k)
            content = ' '.join([item['content'].strip().replace('\n', '') for item in snippets])
            knowledge.append(content)
        return knowledge

    def _get_combined_knowledge(self, abstracts, summaries):
        if self.knowledge_base == 'textbooks':
            knowledge = self._retrieve_knowledge(summaries, "Textbooks")
            return [f"{abs_text}{tb_know}"
                    for abs_text, tb_know in zip(abstracts, knowledge)]
        elif self.knowledge_base == 'statpearls':
            knowledge = self._retrieve_knowledge(summaries, "StatPearls")
            return [f"{abs_text}{sp_know}"
                    for abs_text, sp_know in zip(abstracts, knowledge)]
        else:  # combined
            textbook_knowledge = self._retrieve_knowledge(summaries, "Textbooks")
            statpearl_knowledge = self._retrieve_knowledge(summaries, "StatPearls")
            return [f"{abs_text} {tb_know} {sp_know}"
                    for abs_text, tb_know, sp_know in zip(abstracts, textbook_knowledge, statpearl_knowledge)]

    def evaluate(self, target_sentences: List[str], abstracts: List[str]) -> Dict:
        if len(target_sentences) != len(abstracts):
            raise ValueError("The number of target sentences must match the number of abstracts")

        classifier = self._initialize_classifier()
        self._initialize_bertscore_scorer()
        
        external, external_abs = [], []
        internal, internal_abs = [], []
        
        for abstract, sentence in zip(abstracts, target_sentences):
            label, _ = classifier.predict(abstract, sentence)
            if label == 'yes':
                external.append(sentence)
                external_abs.append(abstract)
            else:
                internal.append(sentence)
                internal_abs.append(abstract)

        external_summaries = [[s] for s in external]
        internal_summaries = [[s] for s in internal]

        generator = self._initialize_answer_extractor()

        if external:
            combined_contexts = self._get_combined_knowledge(external_abs, external_summaries)
            external_results = self.score_batch_qafacteval(
                combined_contexts,
                external_summaries,
                return_qa_pairs=True,
                generator=generator
            )
        else:
            external_results = []

        if internal:
            internal_results = self.score_batch_qafacteval(
                internal_abs,
                internal_summaries,
                return_qa_pairs=True,
                generator=generator
            )
        else:
            internal_results = []

        external_scores = [metrics['qa-eval']['bertscore'] for metrics, *_ in external_results] if external_results else []
        internal_scores = [metrics['qa-eval']['bertscore'] for metrics, *_ in internal_results] if internal_results else []

        return {
            'external_results': external_results,
            'internal_results': internal_results,
            'external_scores': external_scores,
            'internal_scores': internal_scores,
            'external_mean': np.mean(external_scores) if external_scores else 0,
            'internal_mean': np.mean(internal_scores) if internal_scores else 0,
            'overall_mean': np.mean(external_scores + internal_scores) if (external_scores or internal_scores) else 0
        }

    def evaluate_all(self, input_file: str) -> Dict:
        if self.input_file_format == 'csv':
            df = pd.read_csv(input_file, delimiter=self.delimiter, encoding=self.encoding)
        elif self.input_file_format == 'json':
            with open(input_file, 'r', encoding=self.encoding) as f:
                data = json.load(f)
            df = pd.DataFrame(data)
        elif self.input_file_format == 'txt':
            df = pd.read_csv(input_file, delimiter=self.delimiter, encoding=self.encoding)
        else:
            raise ValueError(f"Unsupported file format: {self.input_file_format}")

        classifier = self._initialize_classifier()
        self._initialize_bertscore_scorer()
        
        external, external_abs = [], []
        internal, internal_abs = [], []
        
        for abstract, sentence in zip(df[self.abstract_col], df[self.target_sentence_col]):
            label, _ = classifier.predict(abstract, sentence)
            if label == 'yes':
                external.append(sentence)
                external_abs.append(abstract)
            else:
                internal.append(sentence)
                internal_abs.append(abstract)

        external_summaries = [[s] for s in external]
        internal_summaries = [[s] for s in internal]

        generator = None
        try:
            generator = self._initialize_answer_extractor()
            if generator is None and self.answer_selection_strategy == 'llm-keywords':
                print("Warning: Failed to initialize generator for llm-keywords strategy. Falling back to default strategy.")
                self.answer_selection_strategy = 'none'
        except Exception as e:
            print(f"Error initializing generator: {str(e)}")
            print("Falling back to default strategy.")
            self.answer_selection_strategy = 'none'

        if external:
            combined_contexts = self._get_combined_knowledge(external_abs, external_summaries)
            external_results = self.score_batch_qafacteval(
                combined_contexts,
                external_summaries,
                return_qa_pairs=True,
                generator=generator
            )
        else:
            external_results = []

        if internal:
            internal_results = self.score_batch_qafacteval(
                internal_abs,
                internal_summaries,
                return_qa_pairs=True,
                generator=generator
            )
        else:
            internal_results = []

        external_scores = [metrics['qa-eval']['bertscore'] for metrics, *_ in external_results] if external_results else []
        internal_scores = [metrics['qa-eval']['bertscore'] for metrics, *_ in internal_results] if internal_results else []

        return {
            'external_results': external_results,
            'internal_results': internal_results,
            'external_scores': external_scores,
            'internal_scores': internal_scores,
            'external_mean': np.mean(external_scores) if external_scores else 0,
            'internal_mean': np.mean(internal_scores) if internal_scores else 0,
            'overall_mean': np.mean(external_scores + internal_scores) if (external_scores or internal_scores) else 0
        }

    def score_batch_qafacteval(
        self,
        source: List[SummaryType],
        summaries: List[List[SummaryType]],
        qa_pairs_precomputed: List = None,
        predictions_lists: List = None,
        return_qa_pairs: bool = False,
        generator=None,
            ) -> List[List[MetricsDict]]:
        
        self._initialize_bertscore_scorer()
        
        if generator is None:
            generator = self._initialize_answer_extractor()
        
        if self.answer_selection_strategy == 'llm-keywords' and generator is None:
            raise ValueError("Generator is required for llm-keywords strategy but could not be initialized")

        source = self._flatten_summaries(source)
        summaries = self._flatten_references_list(summaries)

        (
            source,
            summaries,
            is_empty_list,
        ) = self._get_empty_summary_mask(source, summaries)

        if qa_pairs_precomputed:
            qa_pairs_lists = qa_pairs_precomputed
        else:
            qa_pairs_lists = self._generate_qa_pairs(summaries, self.answer_selection_strategy, generator)

        summaries_cons = [x[0] for x in summaries]
        predictions_lists_consistency = self._answer_questions(summaries_cons, qa_pairs_lists)
        qa_pairs_lists_cons = []
        for x, cur_qa_pair in zip(predictions_lists_consistency, qa_pairs_lists):
            qa_summ_new = [[x["prediction"], x["probability"], \
                x["null_probability"]] for x in x[0]]
            answers_ref = [x["answer"] for x in cur_qa_pair[0]]

            bool_total = get_filter(qa_summ_new, answers_ref)

            cur_qa_pair_keep = [x for count, x in enumerate(cur_qa_pair[0]) if bool_total[count]]
            if not cur_qa_pair_keep:
                cur_qa_pair_keep = []
            qa_pairs_lists_cons.append([cur_qa_pair_keep])

        if predictions_lists:
            predictions_lists = predictions_lists
        else:
            predictions_lists = self._answer_questions(source, qa_pairs_lists_cons)
        
        metrics_list, scores_lists = self._score_predictions(
            source, qa_pairs_lists_cons, predictions_lists
        )

        if return_qa_pairs:
            output = self._combine_outputs(
                metrics_list, qa_pairs_lists_cons, predictions_lists, scores_lists
            )
        else:
            output = metrics_list

        output = self._insert_empty_outputs(output, is_empty_list, return_qa_pairs)
        output_final = []
        for out, qa_pairs_list, predictions_cons in \
                zip(output, qa_pairs_lists, predictions_lists_consistency):
            output_final.append((out[0], out[1], qa_pairs_list, predictions_cons))
        return output_final



import hashlib
import logging
from tqdm import tqdm
from typing import Any, Dict, List, Tuple, Union
from qaeval.generation.model import QuestionGenerationModel
from qaeval.answering.model import QuestionAnsweringModel
from qaeval.answer_selection import AnswerSelector
from qaeval.scoring.scorers import (
    ExactMatchScorer,
    F1Scorer,
    IsAnsweredScorer,
    MetaScorer,
)

MetricsDict = Dict[str, float]
SummaryType = Union[str, List[str]]
logger = logging.getLogger(__name__)

class QAEval(object):
    def __init__(
        self,
        generation_model_path: str,
        answering_model_dir: str,
        cuda_device: int = 0,
        generation_batch_size: int = 8,
        answering_batch_size: int = 8,
        verbose: bool = False,
    ) -> None:
        self.question_generator = QuestionGenerationModel(
            generation_model_path,
            cuda_device=cuda_device,
            batch_size=generation_batch_size,
            silent=not verbose,
        )
        self.question_answerer = QuestionAnsweringModel(
            answering_model_dir,
            cuda_device=cuda_device,
            batch_size=answering_batch_size,
            silent=not verbose,
        )
        self.verbose = verbose

        scorers = [IsAnsweredScorer(), ExactMatchScorer(), F1Scorer()]
        self.scorer = MetaScorer(scorers)

    def _flatten_summaries(self, summaries: List[SummaryType]) -> List[str]:
        flat_summaries = []
        for summary in summaries:
            if isinstance(summary, list):
                summary = " ".join(summary)
            flat_summaries.append(summary)
        return flat_summaries

    def _flatten_references_list(
        self, references_list: List[List[SummaryType]]
    ) -> List[List[str]]:
        flat_references_list = []
        for references in references_list:
            flat_references_list.append([])
            for reference in references:
                if isinstance(reference, list):
                    reference = " ".join(reference)
                flat_references_list[-1].append(reference)
        return flat_references_list

    def _get_empty_summary_mask(
        self, summaries: List[str], references_list: List[List[str]]
    ) -> Tuple[List[str], List[List[str]], List[bool]]:
        is_empty_list = []
        non_empty_summaries = []
        non_empty_references_list = []

        for summary, references in zip(summaries, references_list):
            if len(summary.strip()) > 0:
                is_empty_list.append(False)
                non_empty_summaries.append(summary)
                non_empty_references_list.append(references)
            else:
                is_empty_list.append(True)
        return non_empty_summaries, non_empty_references_list, is_empty_list

    def _get_question_id(
        self, instance_index: int, reference_index: int, start: int, end: int
    ) -> str:
        m = hashlib.md5()
        m.update(str(instance_index).encode())
        m.update(str(reference_index).encode())
        m.update(str(start).encode())
        m.update(str(end).encode())
        return m.hexdigest()

    def _generate_qa_pairs(
        self, references_list: List[List[str]], answer_selection_strategy: str, llm_generator: Any = None
    ) -> List[List[List[Dict[str, Any]]]]:
        reference_to_index = {}
        distinct_references_list = []

        mapping = {}
        for i, references in enumerate(references_list):
            for j, reference in enumerate(references):
                if reference not in reference_to_index:
                    reference_to_index[reference] = len(distinct_references_list)
                    distinct_references_list.append(reference)
                mapping[(i, j)] = reference_to_index[reference]

        print(
            f"Selecting answers from {len(distinct_references_list)} distinct summaries"
        )
        if answer_selection_strategy == "llm-keywords" or answer_selection_strategy == "gpt-keywords":
            self.answer_selector = AnswerSelector(answer_selection_strategy, llm_generator)
            answers_list = self.answer_selector.select_all(distinct_references_list)
        else:
            self.answer_selector = AnswerSelector(answer_selection_strategy, generator=None)
            answers_list = self.answer_selector.select_all(distinct_references_list)
        num_answers = sum(len(answers) for answers in answers_list)
        print(f"Selected {num_answers} answers in total")

        generation_inputs = []
        for reference, answers in zip(distinct_references_list, answers_list):
            for answer in answers:
                sentence = reference[answer.sent_start : answer.sent_end]
                start = answer.start - answer.sent_start
                end = answer.end - answer.sent_start
                generation_inputs.append((sentence, start, end))

        print(f"Generating questions for {len(generation_inputs)} answers")
        question_list = self.question_generator.generate_all(generation_inputs)
        print("Finished generating questions")

        index = 0
        remapped_questions = []
        for i, answers in enumerate(answers_list):
            remapped_questions.append([])
            for _ in answers:
                remapped_questions[-1].append(question_list[index])
                index += 1
            assert len(remapped_questions[i]) == len(answers_list[i])
        assert len(remapped_questions) == len(answers_list)

        qa_pairs_lists = []
        for i, references in enumerate(references_list):
            qa_pairs_lists.append([])
            for j, reference in enumerate(references):
                index = mapping[(i, j)]
                qa_pairs_lists[-1].append([])
                for question, answer in zip(
                    remapped_questions[index], answers_list[index]
                ):
                    question_id = self._get_question_id(i, j, answer.start, answer.end)
                    qa_pairs_lists[-1][-1].append(
                        {
                            "question_id": question_id,
                            "question": question,
                            "answer": answer.text,
                            "sent_start": answer.sent_start,
                            "sent_end": answer.sent_end,
                            "answer_start": answer.start,
                            "answer_end": answer.end,
                        }
                    )
        return qa_pairs_lists

    def _get_prediction_id(self, prediction_index: int):
        m = hashlib.md5()
        m.update(str(prediction_index).encode())
        return m.hexdigest()

    def _answer_questions(
        self, summaries: List[str], qa_pairs_lists: List[List[List[Dict[str, Any]]]]
    ) -> List[List[List[Dict[str, Any]]]]:
        qa_inputs = []
        context_to_input_index = {}
        mapping = {}

        for i, (summary, qa_pairs_list) in enumerate(zip(summaries, qa_pairs_lists)):
            for j, qa_pairs in enumerate(qa_pairs_list):
                for k, qa in enumerate(qa_pairs):
                    question = qa["question"]
                    key = (question, summary)
                    if key not in context_to_input_index:
                        context_to_input_index[key] = len(qa_inputs)
                        qa_inputs.append(key)
                    mapping[(i, j, k)] = context_to_input_index[key]
        
        print(f"Answering {len(qa_inputs)} distinct (question, context) pairs")
        predictions = self.question_answerer.answer_all(qa_inputs, return_offsets=True)
        print("Finished answering questions")

        predictions_lists = []
        for i, (summary, qa_pairs_list) in enumerate(zip(summaries, qa_pairs_lists)):
            predictions_lists.append([])
            for j, qa_pairs in enumerate(qa_pairs_list):
                predictions_lists[-1].append([])
                for k, qa in enumerate(qa_pairs):
                    index = mapping[(i, j, k)]
                    prediction, probability, null_probability, offsets = predictions[
                        index
                    ]
                    predictions_lists[-1][-1].append(
                        {
                            "prediction_id": self._get_prediction_id(index),
                            "prediction": prediction,
                            "probability": probability,
                            "null_probability": null_probability,
                            "start": offsets[0],
                            "end": offsets[1],
                        }
                    )
        return predictions_lists

    def _score_predictions(
        self,
        summaries: List[str],
        qa_pairs_lists: List[List[List[Dict[str, Any]]]],
        predictions_lists: List[List[List[Dict[str, Any]]]],
    ) -> Tuple[List[MetricsDict], List[List[List[Dict[str, float]]]]]:
        logger.info("Scoring predictions")
        metrics_list = []
        scores_list = []

        generator = tqdm(
            zip(summaries, qa_pairs_lists, predictions_lists),
            total=len(summaries),
            disable=not self.verbose,
        )
        for summary, qa_pairs_list, predictions_list in generator:
            input_questions_list = []
            input_answers_list = []
            input_predictions_list = []
            input_probabilities_list = []
            input_null_probabilities_list = []
            for qa_pairs, predictions in zip(qa_pairs_list, predictions_list):
                input_questions_list.append([])
                input_answers_list.append([])
                input_predictions_list.append([])
                input_probabilities_list.append([])
                input_null_probabilities_list.append([])
                for qa, prediction in zip(qa_pairs, predictions):
                    input_questions_list[-1].append(qa["question"])
                    input_answers_list[-1].append(qa["answer"])
                    input_predictions_list[-1].append(prediction["prediction"])
                    input_probabilities_list[-1].append(prediction["probability"])
                    input_null_probabilities_list[-1].append(
                        prediction["null_probability"]
                    )

            metrics, scores = self.scorer.score_multi_ref(
                summary,
                input_questions_list,
                input_answers_list,
                input_predictions_list,
                input_probabilities_list,
                input_null_probabilities_list,
            )
            metrics = {"qa-eval": metrics}
            metrics_list.append(metrics)
            scores_list.append(scores)

        logger.info("Finished scoring predictions")
        return metrics_list, scores_list

    def _combine_outputs(
        self,
        metrics_list: List[MetricsDict],
        qa_pairs_lists: List[List[List[Dict[str, Any]]]],
        predictions_lists: List[List[List[Dict[str, Any]]]],
        scores_lists: List[List[List[Dict[str, float]]]],
    ) -> List[List[List[Dict[str, Any]]]]:
        combined = []
        for metrics, qa_pairs_list, predictions_list, scores_list in zip(
            metrics_list, qa_pairs_lists, predictions_lists, scores_lists
        ):
            combined.append((metrics, []))
            for qa_pairs, predictions, scores in zip(
                qa_pairs_list, predictions_list, scores_list
            ):
                combined[-1][1].append([])
                for qa, prediction, score in zip(qa_pairs, predictions, scores):
                    prediction = dict(**prediction)
                    for key in self.scorer.keys():
                        prediction[key] = score[key]
                    combined[-1][1][-1].append(
                        {"question": qa, "prediction": prediction}
                    )
        return combined

    def _insert_empty_outputs(
        self,
        metrics_list: List[MetricsDict],
        is_empty_list: List[bool],
        include_qa_list: bool,
    ) -> List[Any]:
        full_metrics_list = []
        index = 0
        for is_empty in is_empty_list:
            if is_empty:
                empty_metrics = {"qa-eval": self.scorer.default_scores()}
                if include_qa_list:
                    full_metrics_list.append((empty_metrics, []))
                else:
                    full_metrics_list.append(empty_metrics)
            else:
                full_metrics_list.append(metrics_list[index])
                index += 1
        return full_metrics_list

    def score_batch(
        self,
        summaries: List[SummaryType],
        references_list: List[List[SummaryType]],
        return_qa_pairs: bool = False,
    ) -> List[List[MetricsDict]]:
        summaries = self._flatten_summaries(summaries)
        references_list = self._flatten_references_list(references_list)

        (
            summaries,
            references_list,
            is_empty_list,
        ) = self._get_empty_summary_mask(summaries, references_list)

        qa_pairs_lists = self._generate_qa_pairs(references_list)
        predictions_lists = self._answer_questions(summaries, qa_pairs_lists)
        metrics_list, scores_lists = self._score_predictions(
            summaries, qa_pairs_lists, predictions_lists
        )

        if return_qa_pairs:
            output = self._combine_outputs(
                metrics_list, qa_pairs_lists, predictions_lists, scores_lists
            )
        else:
            output = metrics_list
        output = self._insert_empty_outputs(output, is_empty_list, return_qa_pairs)
        return output



from transformers.data.metrics.squad_metrics import compute_f1
from typing import Dict, List, Set
from qaeval.scoring.scorers import Scorer

class F1Scorer(Scorer):
    def keys(self) -> Set[str]:
        return {'f1'}

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        scores = []
        for prediction, answer, prob, null_prob in zip(predictions, answers, probabilities, null_probabilities):
            if prediction is None or null_prob >= prob:
                scores.append({'f1': 0.0})
            else:
                scores.append({'f1': compute_f1(answer, prediction)})
        return scores



import collections
import edlib
from typing import Tuple

from transformers.data.metrics.squad_metrics import (
    get_final_text,
    _get_best_indexes,
    _compute_softmax,
)

def _is_whitespace(c):
    if c == " " or c == "\t" or c == "\r" or c == "\n" or ord(c) == 0x202F:
        return True
    return False


def _get_char_offsets(example, pred_start, pred_end):
    if pred_start is None or pred_end is None:
        return None, None

    token_to_char_start = {}
    token_to_char_end = {}
    for char_index, token_index in enumerate(example.char_to_word_offset):
        if token_index not in token_to_char_start:
            token_to_char_start[token_index] = char_index
        token_to_char_end[token_index] = char_index

    for token_index, end in token_to_char_end.items():
        if token_index == -1:
            # Whitespace at the beginning is mapped to token -1. We don't care about it
            continue
        while _is_whitespace(example.context_text[end]):
            end -= 1
            if end < 0:
                break
        if end < 0:
            raise Exception(f'Token end is less than 0.')
        token_to_char_end[token_index] = end + 1  # exclusive
    return token_to_char_start[pred_start], token_to_char_end[pred_end]


def compute_predictions_logits_with_null(
        tokenizer,
        all_examples,
        all_features,
        all_results,
        n_best_size,
        max_answer_length,
        do_lower_case,
        verbose_logging,
        version_2_with_negative,
        return_offsets = False
):
    example_index_to_features = collections.defaultdict(list)
    for feature in all_features:
        example_index_to_features[feature.example_index].append(feature)

    unique_id_to_result = {}
    for result in all_results:
        unique_id_to_result[result.unique_id] = result

    _PrelimPrediction = collections.namedtuple(  # pylint: disable=invalid-name
        "PrelimPrediction", ["feature_index", "start_index", "end_index", "start_logit", "end_logit"]
    )

    all_predictions = collections.OrderedDict()
    all_nbest_json = collections.OrderedDict()
    scores_diff_json = collections.OrderedDict()
    all_probs = collections.OrderedDict()
    null_scores = collections.OrderedDict()
    offsets = collections.OrderedDict()

    for (example_index, example) in enumerate(all_examples):
        features = example_index_to_features[example_index]

        prelim_predictions = []
        score_null = 1000000
        min_null_feature_index = 0
        null_start_logit = 0
        null_end_logit = 0
        for (feature_index, feature) in enumerate(features):
            result = unique_id_to_result[feature.unique_id]
            start_indexes = _get_best_indexes(result.start_logits, n_best_size)
            end_indexes = _get_best_indexes(result.end_logits, n_best_size)
            if version_2_with_negative:
                feature_null_score = result.start_logits[0] + result.end_logits[0]
                if feature_null_score < score_null:
                    score_null = feature_null_score
                    min_null_feature_index = feature_index
                    null_start_logit = result.start_logits[0]
                    null_end_logit = result.end_logits[0]
            for start_index in start_indexes:
                for end_index in end_indexes:
                    if start_index >= len(feature.tokens):
                        continue
                    if end_index >= len(feature.tokens):
                        continue
                    if start_index not in feature.token_to_orig_map:
                        continue
                    if end_index not in feature.token_to_orig_map:
                        continue
                    if not feature.token_is_max_context.get(start_index, False):
                        continue
                    if end_index < start_index:
                        continue
                    length = end_index - start_index + 1
                    if length > max_answer_length:
                        continue
                    prelim_predictions.append(
                        _PrelimPrediction(
                            feature_index=feature_index,
                            start_index=start_index,
                            end_index=end_index,
                            start_logit=result.start_logits[start_index],
                            end_logit=result.end_logits[end_index],
                        )
                    )
        if version_2_with_negative:
            prelim_predictions.append(
                _PrelimPrediction(
                    feature_index=min_null_feature_index,
                    start_index=0,
                    end_index=0,
                    start_logit=null_start_logit,
                    end_logit=null_end_logit,
                )
            )
        prelim_predictions = sorted(prelim_predictions, key=lambda x: (x.start_logit + x.end_logit), reverse=True)

        _NbestPrediction = collections.namedtuple(  # pylint: disable=invalid-name
            "NbestPrediction", ["text", "start_logit", "end_logit", "doc_start", "doc_end"]
        )

        seen_predictions = {}
        nbest = []
        for pred in prelim_predictions:
            if len(nbest) >= n_best_size:
                break
            feature = features[pred.feature_index]
            if pred.start_index > 0:  # this is a non-null prediction
                tok_tokens = feature.tokens[pred.start_index: (pred.end_index + 1)]
                orig_doc_start = feature.token_to_orig_map[pred.start_index]
                orig_doc_end = feature.token_to_orig_map[pred.end_index]
                orig_tokens = example.doc_tokens[orig_doc_start: (orig_doc_end + 1)]

                tok_text = tokenizer.convert_tokens_to_string(tok_tokens)
                tok_text = tok_text.strip()
                tok_text = " ".join(tok_text.split())
                orig_text = " ".join(orig_tokens)

                final_text = get_final_text(tok_text, orig_text, do_lower_case, verbose_logging)
                if final_text in seen_predictions:
                    continue

                seen_predictions[final_text] = True
            else:
                final_text = ""
                orig_doc_start = None
                orig_doc_end = None
                seen_predictions[final_text] = True

            nbest.append(_NbestPrediction(text=final_text, start_logit=pred.start_logit, end_logit=pred.end_logit,
                                          doc_start=orig_doc_start, doc_end=orig_doc_end))

        if version_2_with_negative:
            if "" not in seen_predictions:
                nbest.append(_NbestPrediction(text="", start_logit=null_start_logit, end_logit=null_end_logit,
                                              doc_start=None, doc_end=None))

            if len(nbest) == 1:
                nbest.insert(0, _NbestPrediction(text="empty", start_logit=0.0, end_logit=0.0, doc_start=None,
                                                 doc_end=None))

        if not nbest:
            nbest.append(_NbestPrediction(text="empty", start_logit=0.0, end_logit=0.0, doc_start=None,
                                          doc_end=None))

        assert len(nbest) >= 1

        total_scores = []
        best_non_null_entry = None
        best_non_null_entry_index = None
        for i, entry in enumerate(nbest):
            total_scores.append(entry.start_logit + entry.end_logit)
            if not best_non_null_entry:
                if entry.text:
                    best_non_null_entry = entry
                    best_non_null_entry_index = i

        probs = _compute_softmax(total_scores)

        nbest_json = []
        null_prob = None
        best_prob = None
        for (i, entry) in enumerate(nbest):
            output = collections.OrderedDict()
            output["text"] = entry.text
            output["probability"] = probs[i]
            output["start_logit"] = entry.start_logit
            output["end_logit"] = entry.end_logit
            if entry.text == '':
                null_prob = probs[i]
            if i == best_non_null_entry_index:
                best_prob = probs[i]
            nbest_json.append(output)

        assert len(nbest_json) >= 1

        if not version_2_with_negative:
            all_predictions[example.qas_id] = nbest_json[0]["text"]
        else:
            # Always predict the best non-null text
            all_predictions[example.qas_id] = best_non_null_entry.text
            all_probs[example.qas_id] = best_prob
            null_scores[example.qas_id] = null_prob
            offsets[example.qas_id] = _get_char_offsets(example, best_non_null_entry.doc_start,
                                                        best_non_null_entry.doc_end)

        all_nbest_json[example.qas_id] = nbest_json

    output = (all_predictions, all_probs, null_scores)
    if return_offsets:
        output = output + (offsets,)
    return output


class SpanFixError(Exception):
    pass


def fix_answer_span(prediction: str, document_span: str, start: int, end: int) -> Tuple[int, int]:
    """
    Tries to fix the answer span of the prediction, which may include some extra whitespace or special characters.

    # Parameters
    - `prediction`: the string output by the QA model
    - `document_span`: the span in the text given by the maybe noisy offsets. See `QuestionAnsweringModel.answer()`
    documentation for more information
    - `start`: the start character offset of `document_span` in the original text
    - `end`: the *exclusive* end character offset of the `document_span` in the original text

    # Returns
    The `start` and *exclusive* `end` character offsets of fixed character offsets of `prediction` in the
    original text.
    """

    if len(prediction) > len(document_span):
        raise SpanFixError(f'Unexpected lengths: "{prediction}", "{document_span}"')

    alignment = edlib.align(prediction, document_span, mode='HW', task='path')
    locations = alignment['locations']
    if len(locations) != 1:
        raise SpanFixError(f'Unable to compute span: "{prediction}", "{document_span}"')
    align_start, align_end = locations[0]

    start += align_start
    end -= len(document_span) - align_end
    end += 1
    return start, end


from typing import Dict, List, Set
import torch
from torch import nn
from transformers import AutoTokenizer, AutoModel
from qaeval.scoring.scorers import Scorer


class LERCModel(nn.Module):
    def __init__(self, pretrained_model_name='bert-large-uncased'):
        super().__init__()
        self.bert = AutoModel.from_pretrained(pretrained_model_name)
        self.score_head = nn.Linear(self.bert.config.hidden_size, 1)
        
    def forward(self, input_ids, attention_mask):
        outputs = self.bert(input_ids=input_ids, attention_mask=attention_mask)
        pooled_output = outputs.last_hidden_state[:, 0, :]  # Use [CLS] token
        score = self.score_head(pooled_output)
        return score.squeeze(-1)


class LERCScorer(Scorer):
    def __init__(self, model_path: str, pretrained_path: str, cuda_device: int, batch_size: int = 8) -> None:
        self.device = torch.device(f"cuda:{cuda_device}" if cuda_device >= 0 else "cpu")
        self.model = torch.load(model_path, map_location=self.device)
        self.model.eval()
        self.tokenizer = AutoTokenizer.from_pretrained('bert-large-uncased') # 'bert-base-uncased'
        self.batch_size = batch_size
        
    def keys(self) -> Set[str]:
        return {'lerc'}
    
    def _prepare_input(self, context: str, question: str, reference: str, candidate: str) -> Dict:
        # Format input following LERC paper
        text = f"Context: {context} Question: {question} Reference: {reference} Candidate: {candidate}"
        encoded = self.tokenizer(
            text,
            max_length=512,
            truncation=True,
            padding='max_length',
            return_tensors='pt'
        )
        return encoded
    
    def _batch_predict(self, batch_inputs: List[Dict]) -> List[float]:
        with torch.no_grad():
            input_ids = torch.cat([x['input_ids'] for x in batch_inputs]).to(self.device)
            attention_mask = torch.cat([x['attention_mask'] for x in batch_inputs]).to(self.device)
            scores = self.model(input_ids, attention_mask)
            return scores.cpu().tolist()

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        input_dicts = []
        indices = []

        for i, (answer, question, prediction, probability, null_probability) in enumerate(
            zip(answers, questions, predictions, probabilities, null_probabilities)
        ):
            encoded_input = self._prepare_input(context, question, answer, prediction)
            input_dicts.append(encoded_input)
            indices.append(i)

        all_scores = []
        for i in range(0, len(input_dicts), self.batch_size):
            batch = input_dicts[i:i + self.batch_size]
            batch_scores = self._batch_predict(batch)
            all_scores.extend(batch_scores)

        scores = [0.0] * len(questions)
        for i, score in zip(indices, all_scores):
            scores[i] = score
        
        return [{'lerc': s} for s in scores]


from typing import Dict, List, Set
from qaeval.scoring.scorers import Scorer

class MetaScorer(Scorer):
    def __init__(self, scorers: List['Scorer']) -> None:
        self.scorers = scorers

    def _merge_dicts(self, dicts: List[Dict[str, float]]) -> Dict[str, float]:
        merged = {}
        for other in dicts:
            merged.update(other)
        return merged

    def keys(self) -> Set[str]:
        keys = set()
        for scorer in self.scorers:
            keys |= scorer.keys()
        return keys

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        scores_list = []
        for scorer in self.scorers:
            _, scores = scorer.score_single_ref(
                context,
                questions,
                answers,
                predictions,
                probabilities,
                null_probabilities
            )
            scores_list.append(scores)

        combined_scores = []
        for i in range(len(questions)):
            combined_scores.append(self._merge_dicts([scores[i] for scores in scores_list]))
        return combined_scores



import os
import gzip
import tqdm
import json

def ends_with_ending_punctuation(s):
    ending_punctuation = ('.', '?', '!')
    return any(s.endswith(char) for char in ending_punctuation)

def concat(title, content):
    if ends_with_ending_punctuation(title.strip()):
        return title.strip() + " " + content.strip()
    else:
        return title.strip() + ". " + content.strip()

def extract(gz_fpath):
    titles = []
    abstracts = []
    title = ""
    abs = ""
    ids = []

    for line in gzip.open(gz_fpath, 'rt').read().split('\n'):
        if line.strip() == "<Article>" or line.strip().startswith("<Article "):
            title = ""
            abs = ""
        elif line.strip() == "</Article>":
            if abs.strip() == "":
                continue
            titles.append(title)
            abstracts.append(abs)
            ids.append(id)
        if line.strip().startswith("<PMID"):
            id = line.strip().strip("</PMID>").split(">")[-1]        
        if line.strip().startswith("<ArticleTitle>"):
            title = line.strip()[14:-15]
        if line.strip().startswith("<AbstractText"):
            if len(abs) == 0: 
                abs += "".join(line.strip()[13:-15].split('>')[1:])
            else:
                abs += " "
                abs += "".join(line.strip()[13:-15].split('>')[1:])

    return titles, abstracts, ids

if __name__ == "__main__":
    fnames = sorted([fname for fname in os.listdir("corpus/pubmed/baseline") if fname.endswith("xml.gz")])
    
    if not os.path.exists("corpus/pubmed/chunk"):
        os.makedirs("corpus/pubmed/chunk")

    for fname in tqdm.tqdm(fnames):
        if os.path.exists("corpus/pubmed/chunk/{:s}".format(fname.replace(".xml.gz", ".jsonl"))):
            continue
        gz_fpath = os.path.join("corpus/pubmed/baseline", fname)
        titles, abstracts, ids = extract(gz_fpath)
        saved_text = [json.dumps({"id": "PMID:"+str(ids[i]), "title": titles[i], "content": abstracts[i], "contents": concat(titles[i], abstracts[i])}) for i in range(len(titles))]
        with open("corpus/pubmed/chunk/{:s}".format(fname.replace(".xml.gz", ".jsonl")), 'w') as f:
            f.write('\n'.join(saved_text))


import os
import tqdm
import json
import regex as re
from datasets import load_dataset
from langchain.text_splitter import RecursiveCharacterTextSplitter

def ends_with_ending_punctuation(s):
    ending_punctuation = ('.', '?', '!')
    return any(s.endswith(char) for char in ending_punctuation)

def concat(title, content):
    if ends_with_ending_punctuation(title.strip()):
        return title.strip() + " " + content.strip()
    else:
        return title.strip() + ". " + content.strip()

if __name__ == "__main__":
    dat = load_dataset("wikipedia", "20220301.en", cache_dir="./corpus/wikipedia", trust_remote_code=True)
    text_splitter = RecursiveCharacterTextSplitter(chunk_size=1000, chunk_overlap=200)

    if not os.path.exists("corpus/wikipedia/chunk"):
        os.makedirs("corpus/wikipedia/chunk")
    
    batch_size = 10000
    len_just = len(str(len(dat['train']) // batch_size + 1))

    saved_text = []
    for i in tqdm.tqdm(range(len(dat['train']))):
        save_id = i // batch_size
        if os.path.exists("corpus/wikipedia/chunk/wiki20220301en{:s}.jsonl".format(str(save_id).rjust(len_just, '0'))):
            continue
        texts = text_splitter.split_text(dat['train'][i]['text'].strip())
        curr_text = [json.dumps({"id": '_'.join([dat['train'][i]['id'], str(j)]), "title": dat['train'][i]['title'], "content": re.sub("\s+", " ", t), "contents": concat(dat['train'][i]['title'], re.sub("\s+", " ", t))}) for j, t in enumerate(texts)]
        saved_text.extend(curr_text)
        if (i + 1) % batch_size == 0:
            with open("corpus/wikipedia/chunk/wiki20220301en{:s}.jsonl".format(str(save_id).rjust(len_just, '0')), 'w') as f:
                f.write('\n'.join(saved_text))
            saved_text = []
    if len(saved_text) > 0:
        with open("corpus/wikipedia/chunk/wiki20220301en{:s}.jsonl".format(str(save_id).rjust(len_just, '0')), 'w') as f:
            f.write('\n'.join(saved_text))


import spacy
import openai
from collections import namedtuple
from spacy.tokens import Span
from typing import List
import configparser
import re
import os
from transformers import pipeline
# config = configparser.ConfigParser()
# config.read('config.ini')

NP_CHUNKS_STRATEGY = 'np-chunks'
MAX_NP_STRATEGY = 'max-np'
NER_STRATEGY = 'ner'
ALL_STRATEGY = 'all'
LLM_KEYWORDS_STRATEGY = 'llm-keywords'
GPT_KEYWORDS_STRATEGY = 'gpt-keywords'

STRATEGIES = [NP_CHUNKS_STRATEGY, MAX_NP_STRATEGY, NER_STRATEGY, ALL_STRATEGY, LLM_KEYWORDS_STRATEGY, GPT_KEYWORDS_STRATEGY]
AnswerOffsets = namedtuple('Answer', ['start', 'end', 'sent_start', 'sent_end', 'text'])

class AnswerSelector(object):
    def __init__(self, strategy: str, generator=None):
        if strategy not in STRATEGIES:
            raise Exception(f'Unknown strategy: {strategy}')
        self.strategy = strategy
        self.nlp = spacy.load('en_core_web_sm')
        if self.strategy == LLM_KEYWORDS_STRATEGY:
            self.generator = generator
        elif self.strategy == GPT_KEYWORDS_STRATEGY:
            self.client = generator

    def _get_np_chunks_answers(self, sentence: Span) -> List[AnswerOffsets]:
        chunks = []
        for chunk in sentence.noun_chunks:
            chunks.append(AnswerOffsets(
                chunk.start_char,
                chunk.end_char,
                sentence.start_char,
                sentence.end_char,
                str(chunk)
            ))
        return chunks

    def _get_max_np_answers(self, sentence: Span) -> List[AnswerOffsets]:
        root = sentence.root
        nodes = [root]
        nps = []

        while nodes:
            node = nodes.pop()
            recurse = True
            if node.pos_ in ['NOUN', 'PROPN']:
                min_index = node.i
                max_index = node.i
                stack = [node]
                while stack:
                    current = stack.pop()
                    min_index = min(min_index, current.i)
                    max_index = max(max_index, current.i)
                    for child in current.children:
                        stack.append(child)

                sent_start_index = sentence[0].i
                num_tokens = max_index - min_index + 1
                if num_tokens <= 7:
                    recurse = False
                    span = sentence[min_index - sent_start_index:max_index + 1 - sent_start_index]
                    nps.append(AnswerOffsets(
                        span.start_char,
                        span.end_char,
                        sentence.start_char,
                        sentence.end_char,
                        str(span)
                    ))

            if recurse:
                for child in node.children:
                    nodes.append(child)

        nps.sort(key=lambda offsets: offsets.start)
        return nps

    def _get_ner_answers(self, sentence: Span) -> List[AnswerOffsets]:
        ners = []
        for entity in sentence.ents:
            if entity.label_ in ['PERSON', 'NORP', 'FAC', 'ORG', 'GPE', 'LOC', 'EVENT', 'WORK_OF_ART']:
                ners.append(AnswerOffsets(
                    entity.start_char,
                    entity.end_char,
                    sentence.start_char,
                    sentence.end_char,
                    str(entity)
                ))
        return ners

    def _get_all_answers(self, sentence: Span) -> List[AnswerOffsets]:
        answers = set()
        answers |= set(self._get_np_chunks_answers(sentence))
        answers |= set(self._get_max_np_answers(sentence))
        answers |= set(self._get_ner_answers(sentence))
        answers = sorted(answers, key=lambda answer: (answer.start, answer.end))
        return answers

    def _call_gpt(self, prompt: str) -> str:
        response = self.client.chat.completions.create(
            model="gpt-4o-2024-11-20",
            messages=[
                {"role": "developer", "content": "QA-based metrics compare information units between the summary and source, so it is thus necessary to first extract such units, or answers, from the given summary. Please extract answers or information units from a plain language summary."},
                {"role": "user", "content": "Extract a comma-separated list of the most important keywords from the following text: \n" + prompt},
            ],
            max_tokens=512,
            temperature=0,
        )
        return response.choices[0].message.content
    
    def _call_llama(self, prompt: str) -> str:
        messages = [
            {"role": "system", "content": "QA-based metrics compare information units between the summary and source, so it is thus necessary to first extract such units, or answers, from the given summary. Please extract answers or information units from a plain language summary."},
            {"role": "user", "content": "Extract a comma-separated list of the most important keywords from the following text: \n" + prompt},
        ]
        outputs = self.generator(messages, do_sample=True, temperature=0.01, max_new_tokens=512)
        generated_text = outputs[0]['generated_text'][-1]['content']
        parts = generated_text.split(":\n\n", 1)
        match = re.search(r"(\b[\w\s\-]+(?:,\s*[\w\s\-]+)+)", generated_text)
        if len(parts) == 2:
            generated_text = parts[1] 
        elif match:
            generated_text = match.group(1)
        else:
            generated_text = generated_text
        return generated_text

    def _get_llmam_keywords_answers(self, text: str) -> List[AnswerOffsets]:
        """
        Prompts the LLM to extract a comma-separated list of keywords from the text.
        Then, for each keyword, finds its first occurrence in the text and determines the sentence boundaries.
        """
        keywords_str = self._call_llama(text)
        keywords = [k.strip() for k in keywords_str.split(",") if k.strip()]
        print(f"Keywords extracted by the LLM: {keywords}")
        answers = []
        for keyword in keywords:
            start = text.find(keyword)
            if start != -1:
                end = start + len(keyword)
                doc = self.nlp(text)
                sent_start = 0
                sent_end = len(text)
                for sent in doc.sents:
                    if start >= sent.start_char and end <= sent.end_char:
                        sent_start = sent.start_char
                        sent_end = sent.end_char
                        break
                answers.append(AnswerOffsets(start, end, sent_start, sent_end, keyword))
        return answers
    
    def _get_gpt_keywords_answers(self, text: str) -> List[AnswerOffsets]:
        """
        Prompts the LLM to extract a comma-separated list of keywords from the text.
        Then, for each keyword, finds its first occurrence in the text and determines the sentence boundaries.
        """
        keywords_str = self._call_gpt(text)
        keywords = [k.strip() for k in keywords_str.split(",") if k.strip()]
        print(f"Keywords extracted by the LLM: {keywords}")
        answers = []
        for keyword in keywords:
            start = text.find(keyword)
            if start != -1:
                end = start + len(keyword)
                doc = self.nlp(text)
                sent_start = 0
                sent_end = len(text)
                for sent in doc.sents:
                    if start >= sent.start_char and end <= sent.end_char:
                        sent_start = sent.start_char
                        sent_end = sent.end_char
                        break
                answers.append(AnswerOffsets(start, end, sent_start, sent_end, keyword))
        return answers

    def select(self, text: str) -> List[AnswerOffsets]:
        """
        Depending on the strategy, returns a list of AnswerOffsets (i.e. extracted keywords or phrases)
        from the input text.
        """
        if self.strategy == LLM_KEYWORDS_STRATEGY:
            return self._get_llmam_keywords_answers(text)
        elif self.strategy == GPT_KEYWORDS_STRATEGY:
            return self._get_gpt_keywords_answers(text)
        else:
            doc = self.nlp(text)
            answers = []
            for sent in doc.sents:
                if self.strategy == NP_CHUNKS_STRATEGY:
                    answers.extend(self._get_np_chunks_answers(sent))
                elif self.strategy == MAX_NP_STRATEGY:
                    answers.extend(self._get_max_np_answers(sent))
                elif self.strategy == NER_STRATEGY:
                    answers.extend(self._get_ner_answers(sent))
                elif self.strategy == ALL_STRATEGY:
                    answers.extend(self._get_all_answers(sent))
                else:
                    raise Exception(f'Unknown strategy: {self.strategy}')
            return answers

    def select_all(self, text_list: List[str]) -> List[List[AnswerOffsets]]:
        return [self.select(text) for text in text_list]



import os
import re
import json
import torch
import transformers
from transformers import AutoTokenizer
import openai
from transformers import StoppingCriteria, StoppingCriteriaList
import tiktoken
import sys
sys.path.append("src")
from src.utils import RetrievalSystem, DocExtracter
from template import *

class MedRAG:

    def __init__(self, cuda_device=0, retrieval=True, follow_up=False, retriever_name="MedCPT", corpus_name="Textbooks", db_dir="./corpus", cache_dir=None, corpus_cache=False, HNSW=False):
        self.retrieval = retrieval
        self.retriever_name = retriever_name
        self.corpus_name = corpus_name
        self.db_dir = db_dir
        self.cache_dir = cache_dir
        self.docExt = None
        self.device = torch.device(f"cuda:{cuda_device}" if torch.cuda.is_available() else "cpu")
        if retrieval:
            self.retrieval_system = RetrievalSystem(self.retriever_name, self.corpus_name, self.db_dir, cache=corpus_cache, HNSW=HNSW, device=self.device)
        else:
            self.retrieval_system = None
        self.templates = {"cot_system": general_cot_system, "cot_prompt": general_cot,
                    "medrag_system": general_medrag_system, "medrag_prompt": general_medrag}
        
        self.follow_up = follow_up
        self.answer = self.medrag_answer

    def medrag_answer(self, question, options=None, k=32, rrf_k=100, save_dir = None, snippets=None, snippets_ids=None, **kwargs):
        '''
        question (str): question to be answered
        options (Dict[str, str]): options to be chosen from
        k (int): number of snippets to retrieve
        rrf_k (int): parameter for Reciprocal Rank Fusion
        save_dir (str): directory to save the results
        snippets (List[Dict]): list of snippets to be used
        snippets_ids (List[Dict]): list of snippet ids to be used
        '''

        if options is not None:
            options = '\n'.join([key+". "+options[key] for key in sorted(options.keys())])
        else:
            options = ''

        # retrieve relevant snippets
        if self.retrieval:
            if snippets is not None:
                retrieved_snippets = snippets[:k]
                scores = []
            elif snippets_ids is not None:
                if self.docExt is None:
                    self.docExt = DocExtracter(db_dir=self.db_dir, cache=True, corpus_name=self.corpus_name)
                retrieved_snippets = self.docExt.extract(snippets_ids[:k])
                scores = []
            else:
                assert self.retrieval_system is not None
                retrieved_snippets, scores = self.retrieval_system.retrieve(question, k=k, rrf_k=rrf_k)
        else:
            retrieved_snippets = []
            scores = []
            contexts = []

        if save_dir is not None and not os.path.exists(save_dir):
            os.makedirs(save_dir)

        return retrieved_snippets, scores

class CustomStoppingCriteria(StoppingCriteria):
    def __init__(self, stop_words, tokenizer, input_len=0):
        super().__init__()
        self.tokenizer = tokenizer
        self.stops_words = stop_words
        self.input_len = input_len

    def __call__(self, input_ids: torch.LongTensor, scores: torch.FloatTensor):
        tokens = self.tokenizer.decode(input_ids[0][self.input_len:])
        return any(stop in tokens for stop in self.stops_words)


from setuptools import setup, find_packages
import os

with open('requirements.txt') as f:
    requirements = [line.strip() for line in f if line.strip() and not line.startswith('#')]

with open('README.md', 'r', encoding='utf-8') as f:
    long_description = f.read()

setup(
    name='plainqafact',
    version='0.12.0',
    author='Zhiwen You',
    author_email='zhiweny2@illinois.edu',
    description='A framework for evaluating plain language summaries using question answering',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/zhiwenyou103/PlainQAFact',
    packages=find_packages(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.9',
    ],
    python_requires='>=3.9',
    install_requires=requirements,
    include_package_data=True,
    package_data={
        'plainqafact': ['transformers_old_tokenizer-3.1.0-py3-none-any.whl'],
    },
    entry_points={
        'console_scripts': [
            'plainqafact=plainqafact.run:main',
        ],
    },
) 


from qaeval.answering.model import QuestionAnsweringModel
from qaeval.answer_selection import AnswerSelector
from qaeval.generation.model import QuestionGenerationModel
from qaeval.metric import QAEval

FIXTURES_ROOT = 'qaeval/tests/fixtures'



from typing import Dict, List, Set
from qaeval.scoring.scorers import Scorer

class IsAnsweredScorer(Scorer):
    def keys(self) -> Set[str]:
        return {'is_answered'}

    def _score_single_ref(
        self,
        context: str,
        questions: List[str],
        answers: List[str],
        predictions: List[str],
        probabilities: List[float],
        null_probabilities: List[float]
    ) -> List[Dict[str, float]]:
        scores = []
        for prob, null_prob in zip(probabilities, null_probabilities):
            if prob > null_prob:
                scores.append({'is_answered': 1.0})
            else:
                scores.append({'is_answered': 0.0})
        return scores
